Package: pkgndep
Type: Package
Title: Analyze Dependency Heaviness of R Packages
Version: 2.0.0
Date: 2026-04-08
Authors@R: person("Zuguang", "Gu", email = "guzuguang@suat-sz.edu.cn", role = c("aut", "cre"),
                  comment = c('ORCID'="0000-0002-7395-8709"))
Depends: R (>= 4.0.0)
Imports: ComplexHeatmap (>= 2.6.0), GetoptLong, GlobalOptions, utils,
        grid, hash, methods, brew
Suggests: knitr, rmarkdown, svglite, base64, testthat
Description: A new metric named 'dependency heaviness' is proposed that measures the number 
    of additional dependency packages that a parent package brings to its child package 
    and are unique to the dependency packages imported by all other parents.  
    The dependency heaviness analysis is visualized by a customized heatmap. 
    The package is described in <doi:10.1093/bioinformatics/btac449>. 
    We have also performed the dependency heaviness analysis on the CRAN/Bioconductor 
    package ecosystem, described in <doi:10.1016/j.jss.2023.111610>. 
URL: https://github.com/jokergoo/pkgndep
VignetteBuilder: knitr
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.1
NeedsCompilation: no
Packaged: 2026-04-08 08:36:17 UTC; guz
Author: Zuguang Gu [aut, cre] (ORCID: <https://orcid.org/0000-0002-7395-8709>)
Maintainer: Zuguang Gu <guzuguang@suat-sz.edu.cn>
Repository: CRAN
Date/Publication: 2026-04-08 09:40:08 UTC
Built: R 4.5.2; ; 2026-04-08 11:01:28 UTC; unix

Package: pkgndep
Type: Package
Title: Analyze Dependency Heaviness of R Packages
Version: 2.0.0
Date: 2026-04-08
Authors@R: person("Zuguang", "Gu", email = "guzuguang@suat-sz.edu.cn", role = c("aut", "cre"),
                  comment = c('ORCID'="0000-0002-7395-8709"))
Depends: R (>= 4.0.0)
Imports: ComplexHeatmap (>= 2.6.0), GetoptLong, GlobalOptions, utils,
        grid, hash, methods, brew
Suggests: knitr, rmarkdown, svglite, base64, testthat
Description: A new metric named 'dependency heaviness' is proposed that measures the number 
    of additional dependency packages that a parent package brings to its child package 
    and are unique to the dependency packages imported by all other parents.  
    The dependency heaviness analysis is visualized by a customized heatmap. 
    The package is described in <doi:10.1093/bioinformatics/btac449>. 
    We have also performed the dependency heaviness analysis on the CRAN/Bioconductor 
    package ecosystem, described in <doi:10.1016/j.jss.2023.111610>. 
URL: https://github.com/jokergoo/pkgndep
VignetteBuilder: knitr
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.1
NeedsCompilation: no
Packaged: 2026-04-08 08:36:17 UTC; guz
Author: Zuguang Gu [aut, cre] (ORCID: <https://orcid.org/0000-0002-7395-8709>)
Maintainer: Zuguang Gu <guzuguang@suat-sz.edu.cn>
Repository: CRAN
Date/Publication: 2026-04-08 09:40:08 UTC
Built: R 4.6.0; ; 2026-04-27 00:14:15 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,pkgndep)
S3method(print,pkgndep)
export(co_heaviness)
export(current_bioc_version)
export(dependency_heatmap)
export(dependency_report)
export(heaviness)
export(heaviness_report)
export(load_pkg_db)
export(pkgndep)
export(pkgndep_opt)
export(reformat_db)
import(ComplexHeatmap)
import(GetoptLong)
import(GlobalOptions)
import(brew)
import(grDevices)
import(grid)
import(methods)
import(utils)

get_bioc_repo = function(version) {
	if(as.numeric_version(version) < as.numeric_version("1.8")) {
		stop("Bioc 'version' should be no earlier than 1.8.")
	}
	version = as.character(version)
	
	c("BioCsoft" = paste0("https://bioconductor.org/packages/", version, "/bioc"),
      "BioCann"  = paste0("https://bioconductor.org/packages/", version, "/data/annotation"),
      "BioCexp"  = paste0("https://bioconductor.org/packages/", version, "/data/experiment")
     )
}

prepare_db = function(lib = .libPaths(), bioc_version = pkgndep_opt$bioc_version, verbose = TRUE) {

	op = getOption("repos")
	on.exit(options(repos = op))
	
	cran_repo = getOption("repos")[["CRAN"]]
	if(!grepl("http", cran_repo, ignore.case = TRUE)) {
		cran_repo = "https://cran.rstudio.com/"
	}
	bioc_repo = get_bioc_repo(bioc_version)
	repos = c(cran_repo, bioc_repo)

	if(verbose) qqcat("retrieve package database from CRAN/Bioconductor (@{bioc_version})...\n")
	suppressMessages(db_remote <- available.packages(repos = repos))

	db_fields = c("Package", "Version", "Depends", "Imports", "LinkingTo", "Suggests", "Enhances", "Repository")
		
	if(identical(lib, NA)) { # only remotes
		db = db_remote[, db_fields]
		db = cbind(db, Local = "no")
		if(verbose) qqcat("  - @{nrow(db)} remote packages on CRAN/Bioconductor.\n")
	} else {
		db_local = installed.packages(lib.loc = lib)
		db_local = cbind(db_local, Repository = NA)
		# db_local = db_local[setdiff(rownames(db_local), BASE_PKGS), , drop = FALSE]

		p1 = db_remote[, "Package"]
		p2 = db_local[, "Package"]
		
		cn = intersect(p1, p2)
		if(length(cn)) {
			db_local[cn, "Repository"] = db_remote[cn, "Repository"]
		}

		db = db_remote[, db_fields]
		
		cn = intersect(p1, p2)
		if(length(cn)) {
			db[cn, ] = db_local[cn, db_fields]
		}

		cn = setdiff(p2, p1)
		if(length(cn)) {
			db = rbind(db, db_local[cn, db_fields])
		}
		db = cbind(db, Local = "no")
		db[p2, "Local"] = "yes"

		if(verbose) qqcat("  - @{nrow(db) - nrow(db_local)} remote packages on CRAN/Bioconductor.\n")
		if(verbose) qqcat("  - @{nrow(db_local)} packages installed locally.\n")
	}

	reformat_db(db)
}

#' Format the package database
#'
#' @param db A data frame returned from [`utils::available.packages()`] or [`utils::installed.packages()`].
#' @param version Version of the database, a self-defined text.
#'
#' @details
#' It reformats the data frame of the package database into a `pkg_db` object.
#'
#' @return
#' A `pkg_db` object. There are the following methods:
#'
#' - `pkg_db$get_meta(package, field = NULL)` ``field`` can take values in "Package", "Version" and "Repository".
#' - `pkg_db$get_dependency_table(package)` Get the dependency table.
#' - `pkg_db$get_rev_dependency_table(package)` Get the reverse dependency table.
#' - `pkg_db$package_dependencies(package, recursive = FALSE, reverse = FALSE, which = "strong", simplify = FALSE)` All the arguments are the same as in [`tools::package_dependencies()`]. Argument `simplify` controls whether to return a data frame or a simplied vector.
#'
#' @export
#' @examples
#' \dontrun{
#' db = available.packages()
#' db2 = reformat_db(db)
#' }
reformat_db = function(db, version = NULL) {

	get_package_list_from_text = function(x) {
		if(is.na(x)) {
			return(character(0))
		}
		x = gsub("\\s*\\(.*?\\)", "", x)
		x = strsplit(x, "\\s*,\\s*")[[1]]
		setdiff(x, c("R", ""))
	}

	if(any(duplicated(db[, "Package"]))) {
		stop_wrap("Package should be unique in the table.")
	}

	n = nrow(db)
	fields = c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")

	cat("prepare dependency table...\n")
	lt = lapply(seq_len(n), function(i) {
		xl = lapply(fields, function(f) get_package_list_from_text(db[i, f]))
		d = unlist(xl)
		cbind(package = rep(db[i, "Package"], length(d)),
		       dependency = d,
		       dep_fields = rep(fields, times = sapply(xl, length)))
	})
	names(lt) = db[, "Package"]

	tb = do.call(rbind, lt)
	rownames(tb) = NULL

	cat("prepare reverse dependency table...\n")
	lt_ind = split(seq_len(nrow(tb)), tb[, "dependency"])
	lt_rev = lapply(lt_ind, function(ind) tb[ind, , drop = FALSE])
	names(lt_rev) = names(lt_ind)

	meta_df = data.frame(Package = db[, "Package"],
			              Version = db[, "Version"])
	if("Local" %in% colnames(db)) {
		meta_df$Local = db[, "Local"]
	}
	if("Repository" %in% colnames(db)) {
		meta_df$Repository = db[, "Repository"]
	}
	rownames(meta_df) = meta_df$Package

	pkg_db(
		meta = meta_df,
		dependency = lt,
		rev_dependency = lt_rev,
		dep_ind_hash = hash::hash(names(lt), seq_along(lt)),
		rev_dep_ind_hash = hash::hash(names(lt_rev), seq_along(lt_rev)),
		version = version
	)
}


pkg_db = setRefClass("pkg_db",
	fields = list(
		"meta" = "data.frame",
		"dependency" = "list",
		"rev_dependency" = "list",
		"dep_ind_hash" = "ANY",
		"rev_dep_ind_hash" = "ANY",
		"version" = "ANY"
	),
	methods = list(
		show = function() {
			qqcat("A package database of @{nrow(.self$meta)} packages.\n")
			if("Local" %in% colnames(.self$meta)) {
				n1 = sum(.self$meta$Local == "yes")
				n2 = sum(.self$meta$Local == "no")
				qqcat("  - @{n1} local / @{n2} remote packages.\n")
			}
			if("Repository" %in% colnames(.self$meta)) {
				n1 = sum(grepl("(bioc|books|annotation|experiment|workflows)/src/contrib$", .self$meta$Repository))
				n2 = sum(grepl("src/contrib$", .self$meta$Repository)) - n1

				if(n1 == 0 && n2 == 0) {
					n1 = sum(.self$meta$Repository == "Bioconductor")
					n2 = sum(.self$meta$Repository == "CRAN")
				}
				n3 = nrow(.self$meta) - n1 - n2
				qqcat("  - @{n2} CRAN / @{n1} Bioconductor / @{n3} other packages.\n")
			}
		})
)

pkg_db$methods(
	get_meta = function(package, field = NULL) {
		if(is.null(field)) {
			.self$meta[package, ]
		} else {
			.self$meta[package, field]
		}
	}
)

pkg_db$methods(
	get_dependency_table = function(package) {
		ind = .self$dep_ind_hash[[package]]
		if(is.null(ind)) {
			return(NULL)
		}
		.self$dependency[[ ind ]]
	}
)

pkg_db$methods(
	get_rev_dependency_table = function(package) {
		ind = .self$rev_dep_ind_hash[[package]]
		if(is.null(ind)) {
			return(NULL)
		}
		.self$rev_dependency[[ ind ]]
	}
)

pkg_db$methods(
	package_dependencies = function(package, recursive = FALSE, reverse = FALSE, 
		which = "strong", simplify = FALSE) {
		
		if(identical(which,"strong")) {
			which = c("Depends", "Imports", "LinkingTo")
		} else if(identical(which, "all")) {
			which = c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")
		} else if (identical(which, "most")) {
			which = c("Depends", "Imports", "LinkingTo", "Suggests")
		}

		empty_tb = matrix(nrow = 0, ncol = 3)
		colnames(empty_tb) = c("package", "dependency", "dep_fields")

		if(reverse) {
			if(recursive) {
				tb = .self$get_rev_dependency_table(package)
				if(is.null(tb)) tb = empty_tb
				tb = tb[tb[, "dep_fields"] %in% which, , drop = FALSE]

				if(nrow(tb) > 0) {
					tbl = list(tb)
					dep_pkg = tb[, "package"]
					while(length(dep_pkg)) {
						tbl2 = lapply(dep_pkg, function(p) {
							tb = .self$get_rev_dependency_table(p)
							if(is.null(tb)) tb = empty_tb
							tb = tb[tb[, "dep_fields"] %in% which, , drop = FALSE]
							tb
						})
						tbl2 = tbl2[sapply(tbl2, nrow) > 0]

						tbl = c(tbl, tbl2)
						dep_pkg = setdiff(unique(unlist(lapply(tbl2, function(x) x[, "package"]))), unlist(lapply(tbl, function(x) x[, "dependency"])))
					}
					tb = do.call(rbind, tbl)
				}	
			} else {
				tb = .self$get_rev_dependency_table(package)
				if(is.null(tb)) tb = empty_tb
				tb = tb[tb[, "dep_fields"] %in% which, , drop = FALSE]
			}

			if(simplify) {
				tb[, "package"][!duplicated(tb[, "package"])]
			} else {
				tb
			}
		} else {
			if(recursive) {
				tb = .self$get_dependency_table(package)
				if(is.null(tb)) tb = empty_tb
				tb = tb[tb[, "dep_fields"] %in% which, , drop = FALSE]

				if(nrow(tb) > 0) {
					tbl = list(tb)
					dep_pkg = tb[, "dependency"]
					while(length(dep_pkg)) {
						tbl2 = lapply(dep_pkg, function(p) {
							tb = .self$get_dependency_table(p)
							if(is.null(tb)) tb = empty_tb
							tb = tb[tb[, "dep_fields"] %in% which, , drop = FALSE]
							tb
						})
						tbl2 = tbl2[sapply(tbl2, nrow) > 0]

						tbl = c(tbl, tbl2)
						dep_pkg = setdiff(unique(unlist(lapply(tbl2, function(x) x[, "dependency"]))), unlist(lapply(tbl, function(x) x[, "package"])))
					}
					tb = do.call(rbind, tbl)
				}				
			} else {
				tb = .self$get_dependency_table(package)
				if(is.null(tb)) tb = empty_tb
				tb = tb[tb[, "dep_fields"] %in% which, , drop = FALSE]
			}

			if(simplify) {
				tb[, "dependency"][!duplicated(tb[, "dependency"])]
			} else {
				tb
			}
		}
	}
)

validate_pkg_db = function(db) {
	all_pkgs = names(db$dependency)

	lt = list()

	all_strong = lapply(db$dependency, function(x) x[x[, "dep_fields"] %in% c("Depends", "Imports", "LinkingTo"), "dependency"])
	all_strong = unique(unlist(all_strong))
	not_in = setdiff(all_strong, all_pkgs)
	lt$strong = not_in

	all = lapply(db$dependency, function(x) x[x[, "dep_fields"] %in% c("Suggests"), "dependency"])
	all = unique(unlist(all))
	not_in = setdiff(all, all_pkgs)
	lt$suggests = not_in

	all = lapply(db$dependency, function(x) x[x[, "dep_fields"] %in% c("Enhances"), "dependency"])
	all = unique(unlist(all))
	not_in = setdiff(all, all_pkgs)
	lt$enhances = not_in

	return(lt)
}

.onAttach = function(libname, pkgname) {
	version = utils::packageDescription(pkgname, fields = "Version")

  	msg = paste0("========================================
", pkgname, " version ", version, "
CRAN page: https://CRAN.R-project.org/package=pkgndep
Github page: https://github.com/jokergoo/pkgndep
Documentation: https://jokergoo.github.io/pkgndep/

If you use it in published research, please cite any of them:
Gu, Z. Pkgndep: a tool for analyzing dependency heaviness 
  of R packages. Bioinformatics 2022.
Gu, Z. On the dependency heaviness of CRAN/Bioconductor 
  ecosystem. Journal of Systems and Software 2023.

This message can be suppressed by:
  suppressPackageStartupMessages(library(pkgndep))
========================================
")	

    packageStartupMessage(msg)

}

#' Current Bioconductor Version
#' 
#' @details
#' The version information is from https://bioconductor.org/config.yaml
#' @export
current_bioc_version = function() {
	ln = readLines("https://bioconductor.org/config.yaml")
	ln = ln[grepl("^release_version", ln)]
	if(length(ln)) {
		bioc_version = gsub('release_version: "(\\d+\\.\\d+)"', "\\1", ln)
	} else {
		bioc_version = ""
	}
	bioc_version
}


.onLoad = function(libname, pkgname) {
	set_var()
	pkgndep_opt$bioc_version = current_bioc_version()
	ENV$lib = .libPaths()
}


set_var = function() {
	installed_tb = utils::installed.packages()
	BASE_PKGS = names(which(installed_tb[ ,"Priority"] == "base", ))
	RECOMMENDED_PKGS = names(which(installed_tb[ ,"Priority"] == "recommended", ))

	assign('BASE_PKGS', BASE_PKGS, envir = topenv())
	assign('RECOMMENDED_PKGS', RECOMMENDED_PKGS, envir = topenv())
}

#' Load package database
#'
#' @param lib Local library. The value can be a character vector specifying the local library paths, or `NULL` which directly uses [`.libPaths()`]. A value of `NA` means not to use local libraries.
#' @param bioc_version Biconductor version.
#' @param verbose Whether to print messages.
#'
#' @details
#' It loads the package database from CRAN/Bioconductor and locally installed packages.
#'
#' The database object internaly is cached for repeated use of other functions in this package.
#'
#' @export
load_pkg_db = function(lib = .libPaths(), bioc_version = pkgndep_opt$bioc_version, verbose = TRUE) {

	ENV$pkg_db = prepare_db(lib = lib, bioc_version = bioc_version, verbose = verbose)
	ENV$bioc_version = bioc_version
	
}

#' Heaviness from parent packages
#'
#' @param x An object returned by [`pkgndep()`].
#' @param rel Whether to return the absolute measure or the relative measure.
#' @param a A constant added for calculating the relative measure.
#' @param only_strong_dep Whether to only return the heaviness for strong parents.
#'
#' @details
#' The heaviness from a parent package is calculated as follows: If package B is in the `Depends`/`Imports`/`LinkingTo` fields of package A,
#' which means, package B is necessary for package A, denote `v1` as the total numbers of packages required for package A,
#' and `v2` as the total number of required packages if moving package B to `Suggests` (which means, now B is not necessary for A).
#' The absolute measure is simply `v1 -  v2` and relative measure is `(v1 + a)/(v2 + a)`. 
#'
#' In the second scenario, if B is in the `Suggests`/`Enhances` fields of package A, now `v2` is the total number of required packages if moving
#' B to `Imports`, the absolute measure is `v2 - v1` and relative measure is `(v2 + a)/(v1 + a)`.
#'
#' @export
#' @examples
#' x = readRDS(system.file("extdata", "ComplexHeatmap_dep.rds", package = "pkgndep"))
#' heaviness(x)
#' heaviness(x, rel = TRUE)
heaviness = function(x, rel = FALSE, a = 10, only_strong_dep = FALSE) {

	if(is.character(x)) {
		stop("`x` should be returned by `pkgndep()`.")
	}

	v1 = length(required_dependency_packages(x, FALSE))
	nr = nrow(x$dep_mat)
	v = numeric(nr)
	for(i in seq_len(nr)) {
		x2 = x

		if(is_field_required(x$dep_fields[i])) {
			x2$dep_fields[i] = "Suggests"
			x2$which_required[i] = FALSE
			x2$which_required_but_not_loaded[i] = FALSE
			v2 = length(required_dependency_packages(x2, FALSE))
			if(rel) {
				v[i] = (v1 + a)/(v2 + a)
			} else {
				v[i] = v1 - v2
			}
		} else {
			x2$dep_fields[i] = "Imports"
			x2$which_required[i] = TRUE
			v2 = length(required_dependency_packages(x2, FALSE))
			if(rel) {
				v[i] = (v2 + a)/(v1 + a)
			} else {
				v[i] = v2 - v1
			}
		}
	}
	names(v) = rownames(x$dep_mat)
	if(only_strong_dep) {
		v[x$which_required]
	} else {
		v
	}
}


heaviness_by_pair = function(x, i, j, rel = FALSE, a = 10, jaccard = FALSE) {
	p = required_dependency_packages(x, FALSE)
	v1 = length(p)
	v = 0
	if(x$which_required[i] & x$which_required[j]) {

		x2 = x
		x2$dep_fields[i] = "Suggests"
		x2$which_required[i] = FALSE
		x2$which_required_but_not_loaded[i] = FALSE
		p_i = setdiff(p, required_dependency_packages(x2, FALSE))

		x2 = x
		x2$dep_fields[j] = "Suggests"
		x2$which_required[j] = FALSE
		x2$which_required_but_not_loaded[j] = FALSE
		p_j = setdiff(p, required_dependency_packages(x2, FALSE))

		x2 = x
		x2$dep_fields[c(i, j)] = "Suggests"
		x2$which_required[c(i, j)] = FALSE
		x2$which_required_but_not_loaded[c(i, j)] = FALSE
		p_ij = setdiff(p, required_dependency_packages(x2, FALSE))

		p_common = setdiff(p_ij, c(p_i, p_j))
		v2 = length(p_common)

		if(jaccard) {
			if(length(p_ij) == 0) {
				v = 0
			} else {
				v = v2/length(p_ij)
			}
		} else if(rel) {
			v = (v1 + a)/((v1 - v2) + a)
		} else {
			v = v2
		}
	}
	v
}

#' Co-heaviness for pairs of parent packages
#'
#' @param x An object returned by [`pkgndep()`].
#' @param rel Whether to return the absolute measure or the relative measure.
#' @param a A constant added for calculating the relative measure.
#' @param jaccard Whether to return Jaccard coeffcient?
#'
#' @details
#' Denote a package as P and its two strong parent packages as A and B, i.e., parent packages in "Depends", "Imports" and "LinkingTo", 
#' the co-heaviness for A and B is calculated as follows.
#'
#' Denote S_A as the set of reduced dependency packages when only moving A to "Suggests" of P, and denote S_B as the set of reduced dependency
#' packages when only moving B to "Suggests" of P, denote S_AB as the set of reduced dependency packages when moving A and B together to "Suggests" of P,
#' the co-heaviness of A, B on P is calculatd as `length(setdiff(S_AB, union(S_A, S_B)))`, which is the number of reduced package only caused by co-action of A and B.
#'
#' Note the co-heaviness is only calculated for parent packages in "Depends", "Imports" and "LinkingTo".
#'
#' When `jaccard` is set to `TRUE`, the function returns jaccard coeffcient. `setdiff(S_AB, union(S_A, S_B))` is actually
#' the set of dependencies imported by and only by two parent packages A and B. Thus the jaccard coeffcient is calculated as
#' `length(setdiff(S_AB, union(S_A, S_B)))/length(S_AB)`.
#'
#' @export
#' @examples
#' \dontrun{
#' # DESeq version 1.36.0, the dependencies have been changed in later versions.
#' x = readRDS(system.file("extdata", "DESeq2_dep.rds", package = "pkgndep"))
#' hm = co_heaviness(x)
#' ComplexHeatmap::Heatmap(hm)
#' co_heaviness(x, jaccard = TRUE)
#' }
co_heaviness = function(x, rel = FALSE, a = 10, jaccard = FALSE) {

	nr = nrow(x$dep_mat)
	m = matrix(NA, nrow = nr, ncol = nr)
	rownames(m) = colnames(m) = rownames(x$dep_mat)
	diag(m) = heaviness(x, rel = rel, a = a)
	if(jaccard) {
		diag(m) = sign(diag(m))
	}

	if(nr <= 1) {
		return(m)
	}

	max_pair = NULL
	v = 0
	for(i in 1:(nr-1)) {
		for(j in (i+1):nr) {
			m[i, j] = m[j, i] = heaviness_by_pair(x, i, j, rel, a, jaccard)
			if(m[i, j] > v) {
				v = m[i, j]
				max_pair = c(i, j)
			}
		}
	}
	if(length(max_pair)) {
		max_pair = rownames(m)[max_pair]
	}
	m2 = m[x$which_required, x$which_required, drop = FALSE]
	if(nrow(m2) > 1) {
		attr(m2, "max") = max(m2[upper.tri(m2)])
		attr(m2, "max_pair") = base::sort(max_pair)
	}
	m2
}

stop_wrap = function(...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    stop(x, call. = FALSE)
}

warning_wrap = function(...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    warning(x, call. = FALSE)
}

message_wrap = function(...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    message(x)
}

gini_index = function(v) {

    if(inherits(v, "pkgndep")) {
        return(gini_index(v$heaviness[v$which_required]))
    }
    if(length(v) < 2) {
        0
    } else {
        v = sort(v)
        n = length(v)
        s = 2 * sum(v * 1:n)/(n * sum(v)) - 1 - 1/n
        s = n/(n - 1)*s
        max(0, s)
    }
}

#' Package dependency analysis
#'
#' @param package Package name. The value can be 1. a CRAN/Bioconductor package, 2. an installed package, 3. a path of a local package, 4. URL of a GitHub repository.
#' @param lib Local library. The value can be a character vector specifying the local library paths, or `NULL` which directly uses [`.libPaths()`]. A value of `NA` means not to use local libraries.
#' @param bioc_version Bioconductor version.
#' @param verbose Whether to show messages.
#'
#' @return
#' A `pkgndep` object.
#' 
#' @export
#'
#' @examples
#' \dontrun{
#' x = pkgndep("ComplexHeatmap")
#' }
#' # The `x` variable generated by `pkgndep()` is already saved in this package.
#' x = readRDS(system.file("extdata", "ComplexHeatmap_dep.rds", package = "pkgndep"))
#' x
#' dependency_heatmap(x)
#' @import GetoptLong
pkgndep = function(package, lib = .libPaths(), bioc_version = pkgndep_opt$bioc_version, verbose = TRUE) {

	# by default, it tests with the newest CRAN/Bioconductor
	# the default Bioconductor is the one corresponding to current R version, but higher bioc version or devel version
	# can be set via pkgndep_opt$bioc_version.

	## check whether package is an CRAN/Bioc package, an installed package, a path or a github link
	if(bioc_version != ENV$bioc_version || !identical(lib, ENV$lib)) {
		load_pkg_db(lib = lib, verbose = verbose, bioc_version = bioc_version)
	}
	pkg_db = ENV$pkg_db

	if(file.exists(package) && !is.null(pkg_db$dep_ind_hash[[package]])) {
		stop(package, "is a local directory and also a package name in the database. If you are analyzing a local directory, change the name to ", package, "/")
	}

	if(is.null(pkg_db$dep_ind_hash[[package]])) {  # pkg is not in the db

		if(pkg_installed(package)) {
			package_name = package
		} else if(file.exists(package)) {
			package_name = basename(package)
		} else if(grepl("^https://github.com/", package)) {
			package_name = basename(package)
		} else {
			stop_wrap("`package` should be 1. a CRAN/Bioconductor package, 2. an installed package,\n3. a path of a local package, 4. URL of a GitHub repository.")
		}
		res = get_package_info_by_path(package)

		version = res$version
		repository = res$repository
		depends = res$depends
		imports = res$imports
		linkingto = res$linkingto
		suggests = res$suggests
		enhances = res$enhances

		imports = setdiff(imports, depends)
		linkingto = setdiff(linkingto, c(imports, depends))
		suggests = setdiff(suggests, c(depends, imports, linkingto))
		enhances = setdiff(enhances, c(depends, imports, linkingto, suggests))

	} else {

		version = pkg_db$get_meta(package)[1, "Version"]
		repository = pkg_db$get_meta(package)[1, "Repository"]

		depends =  pkg_db$package_dependencies(package, which = "Depends", simplify = TRUE)
		imports =  pkg_db$package_dependencies(package, which = "Imports", simplify = TRUE); imports = setdiff(imports, depends)
		linkingto = pkg_db$package_dependencies(package, which = "LinkingTo", simplify = TRUE); linkingto = setdiff(linkingto, c(imports, depends))
		suggests = pkg_db$package_dependencies(package, which = "Suggests", simplify = TRUE); suggests = setdiff(suggests, c(depends, imports, linkingto))
		enhances = pkg_db$package_dependencies(package, which = "Enhances", simplify = TRUE); enhances = setdiff(enhances, c(depends, imports, linkingto, suggests))

		package_name = package
	}

	dep_fields = c(rep("Depends", length(depends)),
		           rep("Imports", length(imports)),
		           rep("LinkingTo", length(linkingto)),
		           rep("Suggests", length(suggests)),
		           rep("Enhances", length(enhances)))
	all_pkgs = c(depends, imports, linkingto, suggests, enhances)

	tb = data.frame(pkg = character(0))	

	if(length(all_pkgs) == 0) {
		obj = list(
			package = package_name,
			version = version,
			repository = repository,
			dep_mat = matrix(nrow = 0, ncol = 0), 
			dep_fields = character(0),

			which_required = logical(0),
			which_required_but_not_loaded = logical(0),
			which_suggested_but_also_loaded = logical(0),
			n_by_strong = 0,
			n_by_all = 0,
			heaviness = numeric(0),
			df_imports = matrix(nrow = 0, ncol = 3, dimnames = list(character(0), c("imports", "importMethods", "importClasses"))),
			pkg_from_session_info = tb$pkg,
			gini_index = 0
		)

		class(obj) = "pkgndep"
		return(obj)
	}

	all_pkgs_dep = lapply(all_pkgs, function(p) {
		pkg_db$package_dependencies(p, which = "strong", recursive = TRUE, simplify = TRUE)
	})

	all_pkgs2 = unique(unlist(all_pkgs_dep))

	rn = all_pkgs
	cn = all_pkgs2

	dep_mat = matrix(0, nrow = length(all_pkgs), ncol = length(all_pkgs2), dimnames = list(rn, cn))
	for(i in seq_along(all_pkgs_dep)) {
		if(length(all_pkgs_dep[[i]])) {
			dep_mat[i, all_pkgs_dep[[i]]] = 1
		}
	}

	df_imports = matrix(0, nrow = nrow(dep_mat), ncol = 3)
	colnames(df_imports) = c("imports", "importMethods", "importClasses")
	rownames(df_imports) = rn

	lt_imports2 = lt_imports = parse_imports_from_namespace(package, package_name)
	if(!is.null(lt_imports)) {
		if(length(lt_imports$n_imports)) {
			lt_imports$n_imports = lt_imports$n_imports[intersect(names(lt_imports$n_imports), rn)]
			df_imports[names(lt_imports$n_imports), "imports"] = lt_imports$n_imports
		}
		if(length(lt_imports$n_import_methods)) {
			lt_imports$n_import_methods = lt_imports$n_import_methods[intersect(names(lt_imports$n_import_methods), rn)]
			df_imports[names(lt_imports$n_import_methods), "importMethods"] = lt_imports$n_import_methods
		}
		if(length(lt_imports$n_import_classes)) {
			lt_imports$n_import_classes = lt_imports$n_import_classes[intersect(names(lt_imports$n_import_classes), rn)]
			df_imports[names(lt_imports$n_import_classes), "importClasses"] = lt_imports$n_import_classes
		}
	}

	not_used = setdiff(rownames(df_imports), c(unlist(lapply(lt_imports, names)), rn[!is_field_required(dep_fields)]))
	if(length(not_used)) {
		df_imports[not_used, 1] = -Inf
	}

	obj = list(
		package = package_name,
		version = version,
		repository = repository,
		dep_mat = dep_mat, 
		dep_fields = dep_fields,

		which_required = is_field_required(dep_fields),
		which_required_but_not_loaded = is.infinite(df_imports[, 1]) & !(rownames(df_imports) %in% tb$pkg),
		which_suggested_but_also_loaded = rn %in% tb$pkg & !is_field_required(dep_fields),
		n_by_strong = 0,
		n_by_all = 0,
		heaviness = 0,
		df_imports = df_imports,
		lt_imports = lt_imports2,
		pkg_from_session_info = tb$pkg
	)

	class(obj) = "pkgndep"

	obj$n_by_strong = length(required_dependency_packages(obj, FALSE))
	obj$n_by_all = length(required_dependency_packages(obj, TRUE))
	obj$heaviness = heaviness(obj)

	row_order = order(factor(dep_fields, levels = c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")), rowSums(dep_mat))

	# a rude way to move all packages which are attached by imported packages before those by suggested packages
	column_order_by = colSums(dep_mat)
	l2 = colSums(dep_mat[is_field_required(dep_fields), ,drop = FALSE]) > 0
	column_order_by[l2] = column_order_by[l2] + 10000
	column_order = order(column_order_by, -apply(dep_mat[row_order, , drop = FALSE], 2, function(x) which(x > 0)[1]), decreasing = TRUE)

	obj$dep_mat = obj$dep_mat[row_order, column_order, drop = FALSE]
	obj$which_required = obj$which_required[row_order]
	obj$which_required_but_not_loaded = obj$which_required_but_not_loaded[row_order]
	obj$which_suggested_but_also_loaded = obj$which_suggested_but_also_loaded[row_order]
	obj$heaviness = obj$heaviness[row_order]
	obj$df_imports = obj$df_imports[row_order, , drop = FALSE]

	obj$gini_index = gini_index(obj$heaviness[obj$which_required])

	return(obj)
}

# internally used, it is approximately 1.5 times faster than the full version of the function
pkgndep_simplified = function(package, pkg_db) {

	depends =  pkg_db$package_dependencies(package, which = "Depends", simplify = TRUE)
	imports =  pkg_db$package_dependencies(package, which = "Imports", simplify = TRUE); imports = setdiff(imports, depends)
	linkingto = pkg_db$package_dependencies(package, which = "LinkingTo", simplify = TRUE); linkingto = setdiff(linkingto, c(imports, depends))
	suggests = pkg_db$package_dependencies(package, which = "Suggests", simplify = TRUE); suggests = setdiff(suggests, c(depends, imports, linkingto))
	enhances = pkg_db$package_dependencies(package, which = "Enhances", simplify = TRUE); enhances = setdiff(enhances, c(depends, imports, linkingto, suggests))

	dep_fields = c(rep("Depends", length(depends)),
		           rep("Imports", length(imports)),
		           rep("LinkingTo", length(linkingto)),
		           rep("Suggests", length(suggests)),
		           rep("Enhances", length(enhances)))
	all_pkgs = c(depends, imports, linkingto, suggests, enhances)

	if(length(all_pkgs) == 0) {
		obj = list(
			package = package,
			version = "0.0.0",
			dep_mat = matrix(nrow = 0, ncol = 0), 
			dep_fields = character(0),

			which_required = logical(0),
			n_by_strong = 0,
			n_by_all = 0,
			heaviness = numeric(0),

			which_required = logical(0),
			which_required_but_not_loaded = logical(0),
			which_suggested_but_also_loaded = logical(0),
			df_imports = matrix(nrow = 0, ncol = 3, dimnames = list(character(0), c("imports", "importMethods", "importClasses")))
		)

		return(obj)
	}

	all_pkgs_dep = lapply(all_pkgs, function(p) {
		pkg_db$package_dependencies(p, which = "strong", recursive = TRUE, simplify = TRUE)
	})

	all_pkgs2 = unique(unlist(all_pkgs_dep))

	rn = all_pkgs
	cn = all_pkgs2

	dep_mat = matrix(0, nrow = length(all_pkgs), ncol = length(all_pkgs2), dimnames = list(rn, cn))
	for(i in seq_along(all_pkgs_dep)) {
		if(length(all_pkgs_dep[[i]])) {
			dep_mat[i, all_pkgs_dep[[i]]] = 1
		}
	}

	obj = list(
		package = package,
		version = "0.0.0",
		dep_mat = dep_mat, 
		dep_fields = dep_fields,

		which_required = is_field_required(dep_fields),
		n_by_strong = 0,
		n_by_all = 0,
		heaviness = 0
	)

	obj$n_by_strong = length(required_dependency_packages(obj, FALSE))
	obj$n_by_all = length(required_dependency_packages(obj, TRUE))
	obj$heaviness = heaviness(obj)

	obj$which_required = logical(nrow(dep_mat))
	obj$which_required_but_not_loaded = logical(nrow(dep_mat))
	obj$which_suggested_but_also_loaded = logical(nrow(dep_mat))

	df_imports = matrix(0, nrow = nrow(dep_mat), ncol = 3)
	colnames(df_imports) = c("imports", "importMethods", "importClasses")
	rownames(df_imports) = rn
	obj$df_imports = df_imports

	class(obj) = "pkgndep"

	obj
}

#' Print method
#'
#' @param x An object from `pkgndep`.
#' @param ... Other arguments.
#'
#' @export
print.pkgndep = function(x, ...) {
	qqcat("'@{x$package}', version @{x$version}\n")
	qqcat("- @{x$n_by_strong} package@{ifelse(x$n_by_strong > 1, 's', '')} @{ifelse(x$n_by_strong > 1, 'are', 'is')} required for installing '@{x$package}'.\n")
	qqcat("- @{x$n_by_all} package@{ifelse(x$n_by_all > 1, 's', '')} @{ifelse(x$n_by_all > 1, 'are', 'is')} required if installing packages listed in all fields in DESCRIPTION.\n")

	if(nrow(x$dep_mat) == 0) {
		return(invisible(NULL))
	}

	l = x$heaviness >= 20 & x$df_imports[, "imports"] > 0 & x$df_imports[, "importMethods"] == 0 & x$df_imports[, "importClasses"] == 0
	if(any(l)) {
		cat("\n")
		cat("Following adjustment could be performed:\n")
		for(i in which(l)) {
			ni = x$df_imports[i, 'imports']
			nm = rownames(x$df_imports)[i]
			qqcat("- Found @{ni} function@{ifelse(ni == 1, ' is', 's are')} imported from a heavy parent '@{nm}'. Moving '@{nm}''\n  to 'Suggests' will reduce @{x$heaviness[i]} dependencies.\n")
		}

		l1 = x$which_required
		l1[l] = FALSE
		m = x$dep_mat
		l2 = colSums(m[l1, , drop = FALSE]) > 0
		n_by_strong2 = length(unique(c(unlist(dimnames(m[l1, l2, drop = FALSE]))))) 
		qqcat("Moving all mentioned packages to 'Suggests' will reduce the dependency packages from @{x$n_by_strong} to @{n_by_strong2}.\n")
	}
	if(!is.null(x$pkgndep_db_version)) {
		qqcat("Using pkgndep_db snapshot, version: @{x$pkgndep_db_version}\n")
	}
}


required_dependency_packages = function(x, all = FALSE) {
	m = x$dep_mat
	if(nrow(m) == 0) {
		return(character(0))
	}
	if(all) {
		unique(c(rownames(m), colnames(m)))
	} else {
		l1 = x$which_required
		l2 = colSums(m[l1, , drop = FALSE]) > 0
		unique(c(unlist(dimnames(m[l1, l2, drop = FALSE]))))
	}
}


is_field_required = function(x) {
	if(inherits(x, "pkgndep")) {
		is_field_required(x$dep_fields)
	} else {
		x %in% c("Depends", "Imports", "LinkingTo")
	}
}


pkg_exists = function(x) {
	system.file(package = x) != ""
}

pkg_installed = pkg_exists


# x can be 1. an installed package, 2. a path (local or github), 3. a cran/bioc package
parse_imports_from_namespace = function(x, pkg = basename(x)) {

	if(pkg %in% names(ENV$ns_data_list)) {
		ns_data = ENV$ns_data_list[[pkg]]
	} else if(pkg_installed(x)) {
		lib_dir = dirname(system.file(package = x))
		if(packageHasNamespace(x, lib_dir)) {
			ns_data = parseNamespaceFile(x, lib_dir)
		} else {
			ns_data = NULL
		}

		ENV$ns_data_list[[pkg]] = ns_data
	} else if(grepl("^https://github.com/", x)) {
		x = gsub("https://github.com/", "https://raw.githubusercontent.com/", x)
		namespace_link = paste0(x, "/master/NAMESPACE")
		description_link = paste0(x, "/master/DESCRIPTION")

		tmpfile1 = tempfile()
		tmpfile2 = tempfile()
		oe1 = try(download.file(namespace_link, tmpfile1, quiet = TRUE), silent = TRUE)
		oe2 = try(download.file(description_link, tmpfile2, quiet = TRUE), silent = TRUE)
		on.exit(file.remove(c(tmpfile1, tmpfile2)))

		if(inherits(oe1, "try-error") || inherits(oe2, "try-error")) {
			ns_data = NULL
		} else {
			ns_data = parseNamespaceFile_cp(x, tmpfile1, tmpfile2)
		}

		ENV$ns_data_list[[pkg]] = ns_data

	} else if(file.exists(x)) {
		namespace_link = paste0(x, "/NAMESPACE")
		description_link = paste0(x, "/DESCRIPTION")

		if(file.exists(namespace_link) && file.exists(description_link)) {
			ns_data = parseNamespaceFile_cp(x, namespace_link, description_link)
		} else {
			ns_data = NULL
		}

		ENV$ns_data_list[[pkg]] = ns_data
	} else {

		load_pkg_db()

		if("Repository" %in% colnames(ENV$pkg_db$meta)) {
			if(any(ENV$pkg_db$meta[x, "Repository"] %in% c("CRAN", "Bioconductor"))) {
				ns_data = NULL
			} else {
				if(grepl("(bioc|books|annotation|experiment|workflow)", ENV$pkg_db$meta[x, "Repository"])) {
					namespace_link = paste0("https://code.bioconductor.org/browse/", x, "/raw/master/NAMESPACE")
					description_link = paste0("https://code.bioconductor.org/browse/", x, "/raw/master/DESCRIPTION")
				} else {
					namespace_link = paste0("https://raw.githubusercontent.com/cran/", x, "/master/NAMESPACE")
					description_link = paste0("https://raw.githubusercontent.com/cran/", x, "/master/DESCRIPTION")
				}

				tmpfile1 = tempfile()
				tmpfile2 = tempfile()
				oe1 = try(download.file(namespace_link, tmpfile1, quiet = TRUE), silent = TRUE)
				oe2 = try(download.file(description_link, tmpfile2, quiet = TRUE), silent = TRUE)
				on.exit({
					if(file.exists(tmpfile1)) file.remove(tmpfile1)
					if(file.exists(tmpfile2)) file.remove(tmpfile2)
				})

				if(inherits(oe1, "try-error") || inherits(oe2, "try-error")) {
					ns_data = NULL
				} else {
					ns_data = parseNamespaceFile_cp(x, tmpfile1, tmpfile2)
				}
			} 
		} else {
			ns_data = NULL
		}

		ENV$ns_data_list[[pkg]] = ns_data
	}

	if(is.null(ns_data)) {
		return(NULL)
	}

	lt = ns_data$imports
	if(length(lt)) {
		lt1 = lt[!sapply(lt, is.list)]
		lt2 = lt[sapply(lt, is.list)]
		n_imports = tapply(seq_along(lt2), unlist(sapply(lt2, function(x) x[1])), 
			function(x) unique(unlist(lapply(lt2[x], function(y) y[2]))))
		import_fun_list = n_imports
		n_imports = sapply(n_imports, length)
		if(length(n_imports)) {
			sign = tapply(seq_along(lt2), unlist(sapply(lt2, function(x) x[1])), 
				function(x) unique(unlist(lapply(lt2[x], function(y) names(y)[2]))))
			sign = sapply(sign, function(x) ifelse(identical(x, "except"), -1, 1))
			n_imports = n_imports * sign
		} else {
		  n_imports = NULL
		}
		n_imports = c(n_imports, structure(rep(0, length(lt1)), names = unlist(lt1)))

		attr(n_imports, "fun_list") = import_fun_list
	} else {
		n_imports = NULL                
	}

	lt = ns_data$importMethods
	if(length(lt)) {
		lt1 = lt[!sapply(lt, is.list)]
		lt2 = lt[sapply(lt, is.list)]
		n_import_methods = tapply(seq_along(lt2), unlist(sapply(lt2, function(x) x[1])), 
			function(x) unique(unlist(lapply(lt2[x], function(y) y[2]))))
		import_method_list = n_import_methods
		n_import_methods = sapply(n_import_methods, length)
		if(length(n_import_methods) == 0) n_import_methods = NULL
		n_import_methods = c(n_import_methods, structure(rep(0, length(lt1)), names = unlist(lt1)))
		attr(n_import_methods, "method_list") = import_method_list
	} else {
		n_import_methods = NULL
	}

	lt = ns_data$importClasses
	if(length(lt)) {
		lt1 = lt[!sapply(lt, is.list)]
		lt2 = lt[sapply(lt, is.list)]
		n_import_classes = tapply(seq_along(lt2), unlist(sapply(lt2, function(x) x[1])), 
			function(x) unique(unlist(lapply(lt2[x], function(y) y[2]))))
		import_class_list = n_import_classes
		n_import_classes = sapply(n_import_classes, length)
		if(length(n_import_classes) == 0) n_import_classes = NULL
		n_import_classes = c(n_import_classes, structure(rep(0, length(lt1)), names = unlist(lt1)))
		attr(n_import_classes, "class_list") = import_class_list
	} else {
		n_import_classes = NULL
	}

	list(n_imports = n_imports, n_import_methods = n_import_methods, n_import_classes = n_import_classes)
}


# modified from base::parseNamespaceFile
parseNamespaceFile_cp <- function(package, nsFile, descfile, mustExist = TRUE) {

    namespaceFilePath <- function(package, package.lib)
        file.path(package.lib, package, "NAMESPACE")

    ## These two functions are essentially local to the parsing of
    ## the namespace file and don't need to be made available to
    ## users.  These manipulate the data from useDynLib() directives
    ## for the same DLL to determine how to map the symbols to R
    ## variables.

    nativeRoutineMap <-
        ## Creates a new NativeRoutineMap.
        function(useRegistration, symbolNames, fixes) {
            proto <- list(useRegistration = FALSE,
                          symbolNames = character())
            class(proto) <- "NativeRoutineMap"

            mergeNativeRoutineMaps(proto, useRegistration, symbolNames, fixes)
        }

    mergeNativeRoutineMaps <-
        ## Merges new settings into a NativeRoutineMap
        function(map, useRegistration, symbolNames, fixes) {
            if(!useRegistration)
                names(symbolNames) <-
                    paste0(fixes[1L],  names(symbolNames), fixes[2L])
            else
                map$registrationFixes <- fixes
            map$useRegistration <- map$useRegistration || useRegistration
            map$symbolNames <- c(map$symbolNames, symbolNames)
            map
        }

     enc <- if (file.exists(descfile)) {
        read.dcf(file = descfile, "Encoding")[1L]
    } else NA_character_
    if (file.exists(nsFile))
        directives <- if (!is.na(enc) &&
                          ! Sys.getlocale("LC_CTYPE") %in% c("C", "POSIX")) {
            lines <- readLines(nsFile, warn = FALSE)
            tmp <- iconv(lines, from = enc, to = "")
            bad <- which(is.na(tmp))
            ## do not report purely comment lines,
            comm <- grep("^[[:space:]]*#", lines[bad],
                         invert = TRUE, useBytes = TRUE)
            if(length(bad[comm]))
                stop("unable to re-encode some lines in NAMESPACE file")
            tmp <- iconv(lines, from = enc, to = "", sub = "byte")
	    con <- textConnection(tmp)
            on.exit(close(con))
	    parse(con, keep.source = FALSE, srcfile = NULL)
        } else parse(nsFile, keep.source = FALSE, srcfile = NULL)
    else if (mustExist)
        stop(gettextf("package %s has no 'NAMESPACE' file", sQuote(package)),
             domain = NA)
    else directives <- NULL
    exports <- character()
    exportPatterns <- character()
    exportClasses <- character()
    exportClassPatterns <- character()
    exportMethods <- character()
    imports <- list()
    importMethods <- list()
    importClasses <- list()
    dynlibs <- character()
    nS3methods <- 1000L
    S3methods <- matrix(NA_character_, nS3methods, 4L)
    nativeRoutines <- list()
    nS3 <- 0L
    parseDirective <- function(e) {
        ## trying to get more helpful error message:
		asChar <- function(cc) {
		    r <- as.character(cc)
		    if(any(r == ""))
			stop(gettextf("empty name in directive '%s' in 'NAMESPACE' file",
				      as.character(e[[1L]])),
			     domain = NA)
		    r
		}
        evalToChar <- function(cc) {
            vars <- all.vars(cc)
            names(vars) <- vars
            as.character(eval(eval(call("substitute", cc, as.list(vars))),
                              .GlobalEnv))
        }
        switch(as.character(e[[1L]]),
               "if" = {
               		oe = try(lll <- eval(e[[2L]], .GlobalEnv), silent = TRUE)
               		if(inherits(oe, "try-error")) {
               			lll = FALSE
               		}
	               if (lll) {
	               		parseDirective(e[[3L]])
	               } else if (length(e) == 4L) {
	               		parseDirective(e[[4L]])
	              	}
	             },
               "{" =  for (ee in as.list(e[-1L])) parseDirective(ee),
               "=" =,
               "<-" = {
                   parseDirective(e[[3L]])
                   if(as.character(e[[3L]][[1L]]) == "useDynLib")
                       names(dynlibs)[length(dynlibs)] <<- asChar(e[[2L]])
               },
               export = {
                   exp <- e[-1L]
                   exp <- structure(asChar(exp), names = names(exp))
                   exports <<- c(exports, exp)
               },
               exportPattern = {
                   pat <- asChar(e[-1L])
                   exportPatterns <<- c(pat, exportPatterns)
               },
               exportClassPattern = {
                   pat <- asChar(e[-1L])
                   exportClassPatterns <<- c(pat, exportClassPatterns)
               },
               exportClass = , exportClasses = {
                   exportClasses <<- c(asChar(e[-1L]), exportClasses)
               },
               exportMethods = {
                   exportMethods <<- c(asChar(e[-1L]), exportMethods)
               },
               import = {
                   except <- e$except
                   e$except <- NULL
                   pkgs <- as.list(asChar(e[-1L]))
                   if (!is.null(except)) {
                       pkgs <- lapply(pkgs, list, except=evalToChar(except))
                   }
                   imports <<- c(imports, pkgs)
               },
               importFrom = {
                   imp <- e[-1L]
                   ivars <- imp[-1L]
                   inames <- names(ivars)
                   imp <- list(asChar(imp[1L]),
                               structure(asChar(ivars), names = inames))
                   imports <<- c(imports, list(imp))
               },
               importClassFrom = , importClassesFrom = {
                   imp <- asChar(e[-1L])
                   pkg <- imp[[1L]]
                   impClasses <- imp[-1L]
                   imp <- list(asChar(pkg), asChar(impClasses))
                   importClasses <<- c(importClasses, list(imp))
               },
               importMethodsFrom = {
                   imp <- asChar(e[-1L])
                   pkg <- imp[[1L]]
                   impMethods <- imp[-1L]
                   imp <- list(asChar(pkg), asChar(impMethods))
                   importMethods <<- c(importMethods, list(imp))
               },
               useDynLib = {

                   ## This attempts to process as much of the
                   ## information as possible when NAMESPACE is parsed
                   ## rather than when it is loaded and creates
                   ## NativeRoutineMap objects to handle the mapping
                   ## of symbols to R variable names.

                   ## The name is the second element after useDynLib
                   dyl <- as.character(e[2L])
                   ## We ensure uniqueness at the end.
                   dynlibs <<-
                       structure(c(dynlibs, dyl),
                                 names = c(names(dynlibs),
                                 ifelse(!is.null(names(e)) &&
                                        nzchar(names(e)[2L]), names(e)[2L], "" )))
                   if (length(e) > 2L) {
                       ## Author has specified some mappings for the symbols

                       symNames <- as.character(e[-c(1L, 2L)])
                       names(symNames) <- names(e[-c(1, 2)])

                       ## If there are no names, then use the names of
                       ## the symbols themselves.
                       if (length(names(symNames)) == 0L)
                           names(symNames) <- symNames
                       else if (any(w <- names(symNames) == "")) {
                           names(symNames)[w] <- symNames[w]
                       }

                       ## For each DLL, we build up a list the (R
                       ## variable name, symbol name) mappings. We do
                       ## this in a NativeRoutineMap object and we
                       ## merge potentially multiple useDynLib()
                       ## directives for the same DLL into a single
                       ## map.  Then we have separate NativeRoutineMap
                       ## for each different DLL.  E.g. if we have
                       ## useDynLib(foo, a, b, c) and useDynLib(bar,
                       ## a, x, y) we would maintain and resolve them
                       ## separately.

                       dup <- duplicated(names(symNames))
                       if (any(dup))
                           warning(gettextf("duplicate symbol names %s in useDynLib(\"%s\")",
                                            paste(sQuote(names(symNames)[dup]),
                                                  collapse = ", "), dyl),
                                   domain = NA, call. = FALSE)

                       symNames <- symNames[!dup]

                       ## Deal with any prefix/suffix pair.
                       fixes <- c("", "")
                       idx <- match(".fixes", names(symNames))
                       if(!is.na(idx)) {
                           ## Take .fixes and treat it as a call,
                           ## e.g. c("pre", "post") or a regular name
                           ## as the prefix.
                           if(nzchar(symNames[idx])) {
                               e <- parse(text = symNames[idx],
                                          keep.source = FALSE,
                                          srcfile = NULL)[[1L]]
                               if(is.call(e))
                                   val <- eval(e, .GlobalEnv)
                               else
                                   val <- as.character(e)
                               if(length(val))
                                   fixes[seq_along(val)] <- val
                           }
                           symNames <- symNames[-idx]
                       }

                       ## Deal with a .registration entry. It must be
                       ## .registration = value and value will be coerced
                       ## to a logical.
                       useRegistration <- FALSE
                       idx <- match(".registration", names(symNames))
                       if(!is.na(idx)) {
                           useRegistration <- as.logical(symNames[idx])
                           symNames <- symNames[-idx]
                       }

                       ## Now merge into the NativeRoutineMap.
                       nativeRoutines[[ dyl ]] <<-
                           if(dyl %in% names(nativeRoutines))
                               mergeNativeRoutineMaps(nativeRoutines[[ dyl ]],
                                                      useRegistration,
                                                      symNames, fixes)
                           else
                               nativeRoutineMap(useRegistration, symNames,
                                                fixes)
                   }
               },
               S3method = {
                   spec <- e[-1L]
                   if (length(spec) != 2L && length(spec) != 3L)
                       stop(gettextf("bad 'S3method' directive: %s",
                                     deparse(e)),
                            call. = FALSE, domain = NA)
                   nS3 <<- nS3 + 1L
                   if(nS3 > nS3methods) {
                       old <- S3methods
                       nold <- nS3methods
                       nS3methods <<- nS3methods * 2L
                       new <- matrix(NA_character_, nS3methods, 4L)
                       ind <- seq_len(nold)
                       for (i in 1:4) new[ind, i] <- old[ind, i]
                       S3methods <<- new
                       rm(old, new)
                   }
                   if(is.call(gen <- spec[[1L]]) &&
                      identical(as.character(gen[[1L]]), "::")) {
                       pkg <- as.character(gen[[2L]])[1L]
                       gen <- as.character(gen[[3L]])[1L]
                       S3methods[nS3, c(seq_along(spec), 4L)] <<-
                           c(gen, asChar(spec[-1L]), pkg)
                   } else
                   S3methods[nS3, seq_along(spec)] <<- asChar(spec)
               },
               stop(gettextf("unknown namespace directive: %s", deparse(e, nlines=1L)),
                    call. = FALSE, domain = NA)
               )
    }

    for (e in directives)
        parseDirective(e)

    ## need to preserve the names on dynlibs, so unique() is not appropriate.
    dynlibs <- dynlibs[!duplicated(dynlibs)]
    list(imports = imports, exports = exports,
         exportPatterns = unique(exportPatterns),
         importClasses = importClasses, importMethods = importMethods,
         exportClasses = unique(exportClasses),
         exportMethods = unique(exportMethods),
         exportClassPatterns = unique(exportClassPatterns),
         dynlibs = dynlibs, nativeRoutines = nativeRoutines,
         S3methods = unique(S3methods[seq_len(nS3), , drop = FALSE]) )
} ## end{parseNamespaceFile}

get_package_info_by_path = function(path) {
	if(grepl("^https://github.com/", path)) {
		dcf = tempfile()
		path = gsub("https://github.com/", "https://raw.githubusercontent.com/", path)
		download.file(paste0(path, "/master/DESCRIPTION"), dcf, quiet = TRUE)
		on.exit(file.remove(dcf))
	} else if(pkg_installed(path)) {
		dcf = system.file("DESCRIPTION", package = path)
	} else {
		dcf = paste0(path, "/", "DESCRIPTION")
	}
	de = read.dcf(dcf)
	if("Depends" %in% colnames(de)) {
		depends = de[1, "Depends"]
		depends = gsub("\\(.*?\\)", "", depends)
		depends = gsub("\\s", "", depends)
		depends = strsplit(depends, ",")[[1]]
		depends = setdiff(depends, "R")
	} else {
		depends = character(0)
	}

	if("Imports" %in% colnames(de)) {
		imports = de[1, "Imports"]
		imports = gsub("\\(.*?\\)", "", imports)
		imports = gsub("\\s", "", imports)
		imports = strsplit(imports, ",")[[1]]
	} else {
		imports = character(0)
	}

	if("Suggests" %in% colnames(de)) {
		suggests = de[1, "Suggests"]
		suggests = gsub("\\(.*?\\)", "", suggests)
		suggests = gsub("\\s", "", suggests)
		suggests = strsplit(suggests, ",")[[1]]
	} else {
		suggests = character(0)
	}

	if("LinkingTo" %in% colnames(de)) {
		linkingto = de[1, "LinkingTo"]
		linkingto = gsub("\\(.*?\\)", "", linkingto)
		linkingto = gsub("\\s", "", linkingto)
		linkingto = strsplit(linkingto, ",")[[1]]
	} else {
		linkingto = character(0)
	}

	if("Enhances" %in% colnames(de)) {
		enhances = de[1, "Enhances"]
		enhances = gsub("\\(.*?\\)", "", enhances)
		enhances = gsub("\\s", "", enhances)
		enhances = strsplit(enhances, ",")[[1]]
	} else {
		enhances = character(0)
	}

	version = de[1, "Version"]
	
	if(grepl("^https://", path)) {
		repository = "GitHub"
	} else {
		repository = "local"
	}

	list(depends = depends, imports = imports, linkingto = linkingto, suggests = suggests, enhances = enhances, 
		version = version, repository = repository)

}

#' HTML report for package dependency heaviness analysis
#'
#' @param pkg An object from [`pkgndep()`].
#' @param file The path of the html file. If it is not specified, the report will be automatically opened in the web browser.
#'
#' @return
#' The path of the HTML file of the report.
#' @rdname report
#' @export
#' @examples
#' if(interactive()) {
#'     x = readRDS(system.file("extdata", "ComplexHeatmap_dep.rds", package = "pkgndep"))
#'     heaviness_report(x)
#' }
#' @import brew
heaviness_report = function(pkg, file = NULL) {
	
	template_file = system.file("extdata", "package_report.html", package = "pkgndep")

	if(!requireNamespace("base64", quietly = TRUE)) {
		stop("Please install base64 package.")
	}

	open_browser = FALSE
	if(is.null(file)) {
		file = tempfile(fileext = ".html")
		open_browser = TRUE
	}
	
	tb = NULL
	n_total = nrow(pkg$df_imports)

	if(n_total > 0) {

		required_pkgs = rowSums(pkg$dep_mat)
		row_order = order(factor(pkg$dep_fields, levels = FIELDS), -pkg$heaviness)
		tb = as.data.frame(pkg$df_imports)
		tb = cbind(field = pkg$dep_fields, tb)
		tb$required_pkgs = required_pkgs
		tb$heaviness = pkg$heaviness

		tb = tb[row_order, , drop = FALSE]
		tb = cbind("Package"= rownames(tb), tb)
		tb = as.matrix(tb)
	}

	brew::brew(template_file, file)
	
	if(open_browser) {
		browseURL(file)
	}

	invisible(file)
}


#' @rdname report
#' @export
#' @param ... pass to [`heaviness_report()`].
dependency_report = function(...) {
	heaviness_report(...)
}

ENV = new.env()

FIELDS = c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")

ENV$bioc_version = "not_configured"

#' Global parameters for pkgndep
#'
#' @param ... Arguments for the parameters, see "details" section
#' @param RESET Reset to default values.
#' @param READ.ONLY Please ignore.
#' @param LOCAL Pllease ignore.
#' @param ADD Please ignore.
#' 
#' @details
#' There are following parameters:
#' 
#' - `bioc_version` The bioconductor version. Use numeric version for devel branch.
#' 
#' @export
#' @import GlobalOptions
pkgndep_opt = setGlobalOptions(
    tmp_dir = list(.value = ".", .visible = FALSE),
    bioc_version = ENV$bioc_version
)

#' Make the dependency heatmap
#'
#' @param x An object from [`pkgndep()`].
#' @param ... Other arguments.
#'
#' @rdname plot
#' @export
#' @import grid
#' @import ComplexHeatmap
plot.pkgndep = function(x, ...) {
	dependency_heatmap(x, ...)
}

#' @rdname plot
#' @export
#'
#' @param pkg_fontsize Font size for the package names.
#' @param title_fontsize Font size for the title.
#' @param legend_fontsize Font size for the legends.
#' @param fix_size Should the rows and columns in the heatmap have fixed size?
#' @param cex A factor multiplicated to all font sizes.
#' @param help Whether to print help message?
#' @param file A path of the figure. The size of the figure is automatically calculated.
#' @param res Resolution of the figure (only for png and jpeg).
#'
#' @details
#' If `fix_size` is set to `TRUE`. The size of the whole plot can be obtained by:
#'
#' ```
#' size = dependency_heatmap(x, fix_size = TRUE)
#' ```
#'
#' where `size` is a numeric vector of length two which are the width and height of the whole heatmap.
#' If `file` argument is set, the size of the figure is automatically calculated.
#'
#' If there are no dependency packages stored in `x`, `NULL` is returned.
#' 
#' @return
#' A vector of two numeric values (in inches) that correspond to the width and height of the plot.
#' @import methods
#' @import utils
#' @import grDevices
dependency_heatmap = function(x, pkg_fontsize = 10*cex, title_fontsize = 12*cex, 
	legend_fontsize = 10*cex, fix_size = !dev.interactive(), cex = 1, 
	help = TRUE, file = NULL, res = 144) {

	if(!is.null(file)) {
		if(grepl("\\.png$", file, ignore.case = TRUE)) {
			tmp_file = tempfile(fileext = ".png")
			png(tmp_file, width = 1000)
			size = dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
				legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
			dev.off()
			file.remove(tmp_file)

			if(!is.null(size)) {
				size[1] = max(size[1], 7)
				size[2] = max(size[2], 3)
				png(file, width = size[1], height = size[2], units = "in", res = res)
				dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
					legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
				dev.off()
			} else {
				png(file, width = 5, height = 1, units = "in", res = res)
				if(nrow(x$dep_mat) > 0 && ncol(x$dep_mat) == 0) {
					grid.text(qq("All required packages for '@{x$package}' are base packages."), 0.5, 0.5)
				} else {
					grid.text(qq("No dependency found for package '@{x$package}'"), 0.5, 0.5)
				}
				dev.off()
			}

			return(invisible(NULL))
		} else if(grepl("\\.(jpg|jpeg)$", file, ignore.case = TRUE)) {
			tmp_file = tempfile()
			jpeg(tmp_file)
			size = dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
				legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
			dev.off()
			file.remove(tmp_file)

			if(!is.null(size)) {
				size[1] = max(size[1], 7)
				size[2] = max(size[2], 3)
				jpeg(file, width = size[1], height = size[2], units = "in", res = res)
				dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
					legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
				dev.off()
			} else {
				jpeg(file, width = 5, height = 1, units = "in", res = res)
				if(nrow(x$dep_mat) > 0 && ncol(x$dep_mat) == 0) {
					grid.text(qq("All required packages for '@{x$package}' are base packages."), 0.5, 0.5)
				} else {
					grid.text(qq("No dependency found for package '@{x$package}'"), 0.5, 0.5)
				}
				dev.off()
			}

			return(invisible(NULL))
		} else if(grepl("\\.pdf$", file, ignore.case = TRUE)) {
			pdf(NULL)
			size = dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
				legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
			dev.off()

			if(!is.null(size)) {
				size[1] = max(size[1], 7)
				size[2] = max(size[2], 3)
				pdf(file, width = size[1], height = size[2])
				dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
					legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
				dev.off()
			} else {
				pdf(file, width = 5, height = 1)
				if(nrow(x$dep_mat) > 0 && ncol(x$dep_mat) == 0) {
					grid.text(qq("All required packages for '@{x$package}' are base packages."), 0.5, 0.5)
				} else {
					grid.text(qq("No dependency found for package '@{x$package}'"), 0.5, 0.5)
				}
				dev.off()
			}

			return(invisible(NULL))
		} else if(grepl("\\.svg$", file, ignore.case = TRUE)) {
			tmp_file = tempfile()
			svg(tmp_file)
			size = dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
				legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
			dev.off()
			file.remove(tmp_file)

			if(requireNamespace("svglite", quietly = TRUE)) {
				svg_fun = svglite::svglite
			} else {
				svg_fun = svg
			}

			if(!is.null(size)) {
				size[1] = max(size[1], 7)
				size[2] = max(size[2], 3)
				svg_fun(file, width = size[1], height = size[2])
				dependency_heatmap(x, fix_size = TRUE, pkg_fontsize = pkg_fontsize, title_fontsize = title_fontsize,
					legend_fontsize = legend_fontsize, cex = cex, help = FALSE, file = NULL)
				dev.off()
			} else {
				svg_fun(file, width = 5, height = 1)
				if(nrow(x$dep_mat) > 0 && ncol(x$dep_mat) == 0) {
					grid.text(qq("All required packages for '@{x$package}' are base packages."), 0.5, 0.5)
				} else {
					grid.text(qq("No dependency found for package '@{x$package}'"), 0.5, 0.5)
				}
				dev.off()
			}

			return(invisible(NULL))
		} else {
			stop("`file` only allows extensions of 'png/jpg/svg/pdf'.")
		}
	}

	#get.gpar() # test whether the graphics window can be opened or not
	if(missing(fix_size) && cex != 1) {
		fix_size = TRUE
	}

	m = x$dep_mat
	row_split = x$dep_fields

	if(ncol(m) == 0) {
		if(help) message(qq("No dependency found for package '@{x$package}'. Won't generate the plot.\n"))
		return(invisible(NULL))
	}

	line_height = grobHeight(textGrob("A", gp = gpar(fontsize = pkg_fontsize)))*1.5

	fix_size = fix_size
	
	m2 = m
	m2[, colnames(m) %in% BASE_PKGS] = m2[, colnames(m) %in% BASE_PKGS] * 3
	m2[, colnames(m) %in% RECOMMENDED_PKGS] = m2[, colnames(m) %in% RECOMMENDED_PKGS] * 2

    dep_fields_col = structure(c("#d9534f", "#f0ad4e", "#337ab7", "#5cb85c", "#5bc0de"), names =  c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances"))

	ht = Heatmap(m2, 
		name = x$package,
		row_split = factor(row_split, levels = c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")),
		left_annotation = rowAnnotation(dep_fields = x$dep_fields, simple_anno_size = unit(2, "mm"),
			col = list(dep_fields = dep_fields_col),
			show_annotation_name = FALSE, annotation_legend_param = list(dep_fields = list(title = "", at = intersect(names(dep_fields_col), as.character(row_split))))),
		column_split = factor(ifelse(colnames(m) %in% BASE_PKGS, "Base packages", "Other packages"), levels = c("Base packages", "Other packages")),
		rect_gp = gpar(col = "#EEEEEE"),
		show_row_dend = FALSE, 
		show_column_dend = FALSE,
		col = c("1" = "#1f78b4", "2" = "darkgreen", "3" = "#e31a1c", "0" = "#CCCCCC"),
		column_names_gp = gpar(fontsize = pkg_fontsize),
		column_names_rot = 60,
		row_names_gp = gpar(fontsize = pkg_fontsize),
		column_title_gp = gpar(fontsize = title_fontsize),
		row_title_gp = gpar(fontsize = title_fontsize),
		row_title_rot = 90,
		cluster_rows = FALSE,
		cluster_columns = FALSE,
		bottom_annotation = HeatmapAnnotation(
			required1 = anno_simple(
				ifelse(colnames(m) %in% required_dependency_packages(x) & colSums(m[x$which_required, , drop = FALSE]) > 0, "yes", "no"),
				col = c("yes" = "purple", "no" = "white"),
				height = unit(1, "mm")
			), 
			loaded1 = if(length(x$pkg_from_session_info) == 0) NULL else anno_simple(
				ifelse(colnames(m) %in% x$pkg_from_session_info, "yes", "no"),
				col = c("yes" = "white", "no" = "white"),
				pch = ifelse(colnames(m) %in% x$pkg_from_session_info, 16, NA), 
				pt_size = unit(2, "mm"), pt_gp = gpar(col = "purple"),
				height = unit(2, "mm")
			),
			show_annotation_name = FALSE),
		width = if(fix_size) ncol(m)*line_height else NULL,
		height = if(fix_size) nrow(m)*line_height else NULL,
		right_annotation = if(any(is_field_required(x$dep_fields))) {
			rowAnnotation(
				required2 = anno_simple(
					ifelse(rownames(m) %in% required_dependency_packages(x) & is_field_required(x$dep_fields), "yes", "no"),
					col = c("yes" = "purple", "no" = "white"),
					width = unit(1, "mm")
				), 
				loaded2 = if(length(x$pkg_from_session_info) == 0) NULL else anno_simple(
					ifelse(rownames(m) %in% x$pkg_from_session_info, "yes", "no"),
					col = c("yes" = "white", "no" = "white"),
					pch = ifelse(rownames(m) %in% x$pkg_from_session_info, 16, NA), 
					pt_size = unit(2, "mm"), pt_gp = gpar(col = "purple"),
					width = unit(2, "mm")
				),
				show_annotation_name = FALSE)
		} else {
			NULL
		},
		show_heatmap_legend = FALSE
	)

	graphics = list(
        function(x, y, w, h) grid.rect(x, y, w, h, gp = gpar(fill = "#e31a1c", col = "white")),
        function(x, y, w, h) grid.rect(x, y, w, h, gp = gpar(fill = "darkgreen", col = "white")),
        function(x, y, w, h) grid.rect(x, y, w, h, gp = gpar(fill = "#1f78b4", col = "white")),
        function(x, y, w, h) grid.rect(x, y, w, unit(1.6, "mm"), gp = gpar(fill = "purple", col = "white")),
        function(x, y, w, h) grid.points(x, y, pch = 16, size = unit(2, "mm"), gp = gpar(col = "purple"))
    )
    legend_labels = c("Base packages that are required", 
    	              "Recommended packages that are required", 
    	              "Contributed packages that are required", 
    	              qq("Packages that are required for installing '@{x$package}'"),
    	              qq("Packages that are loaded after 'library(@{x$package})'"))

    graphics_ind = seq_along(graphics)

    if(!any(colnames(m2) %in% RECOMMENDED_PKGS)) {
    	graphics_ind = setdiff(graphics_ind, 2)
    }

    if(length(x$pkg_from_session_info) == 0) {
    	graphics_ind = graphics_ind[-length(graphics_ind)]
    }

    graphics = graphics[graphics_ind]
    legend_labels = legend_labels[graphics_ind]
	
	# for the lower version of ComplexHeatmap
    if(packageVersion("ComplexHeatmap") > "2.7.1") {
    	lgd0 = Legend(labels = legend_labels,
			graphics = graphics,
			title = "", labels_gp = gpar(fontsize = legend_fontsize))
		lgd_list = list(lgd0)
	} else {
		lgd_list = NULL
	}
	
	df_imports = x$df_imports
	if(any(is.na(df_imports[, 1]))) {
		parse_namespace = FALSE
	} else {
		parse_namespace = TRUE
	}

	if(parse_namespace) {
		if(any(x$which_suggested_but_also_loaded)) {
			df_imports[x$which_suggsted_but_also_loaded, 1] = -Inf
		}
		ht = ht + rowAnnotation("n_import" = anno_nimports_barplot(df_imports, x$dep_fields, width = unit(1, "cm"),
				gp = gpar(fill = c("#ff7f00", "#cab2d6", "#8dd3c7"), col = NA),
				axis_param = list(gp = gpar(fontsize = 8*cex))),
				annotation_label = "Imported\nmethods", annotation_name_rot = 60,
				annotation_name_gp = gpar(fontsize = pkg_fontsize),
				annotation_name_offset = unit(8, "mm"))
	}

	ht = ht + rowAnnotation(n_pkg = anno_barplot(rowSums(x$dep_mat), width = unit(1, "cm"),
								axis_param = list(gp = gpar(fontsize = 8*cex)), gp = gpar(fill = "#808080", col = NA)),
			annotation_label = "Required\npackages", annotation_name_rot = 60,
			annotation_name_gp = gpar(fontsize = pkg_fontsize),
			annotation_name_offset = unit(8, "mm"))

	ht = ht + rowAnnotation(heaviness = anno_barplot(x$heaviness, width = unit(1, "cm"), ylim = if(all(x$heaviness == 0)) c(0, 1) else NULL,
								axis_param = list(gp = gpar(fontsize = 8*cex)), gp = gpar(fill = "#808080", col = NA)),
			annotation_label = "Heaviness from \nparents", annotation_name_rot = 60,
			annotation_name_gp = gpar(fontsize = pkg_fontsize),
			annotation_name_offset = unit(8, "mm"))
	
	if(parse_namespace) {
		lgd1 = NULL
		if(any(rowSums(df_imports) > 0)) {
			if(all(df_imports[, 1] == 0 & df_imports[, 2] == 0)) {
				ind = 3
			} else if(all(df_imports[, 1] == 0 & df_imports[, 3] == 0)) {
				ind = 2
			} else if(all(df_imports[, 2] == 0 & df_imports[, 3] == 0)) {
				ind = 1
			} else if(all(df_imports[, 1] == 0)) {
				ind = 2:3
			} else if(all(df_imports[, 2] == 0)) {
				ind = c(1, 3)
			} else if(all(df_imports[, 3] == 0)) {
				ind = 1:2
			} else {
				ind = 1:3
			}
			lgd1 = Legend(title = "", at = c("Imported functions", "Imported S4 methods", "Imported S4 classes")[ind], 
				graphics = list(
					function(x, y, w, h) grid.rect(x, y, w, unit(1.6, "mm"), gp = gpar(fill = "#ff7f00", col = "white")),
					function(x, y, w, h) grid.rect(x, y, w, unit(1.6, "mm"), gp = gpar(fill = "#cab2d6", col = "white")),
					function(x, y, w, h) grid.rect(x, y, w, unit(1.6, "mm"), gp = gpar(fill = "#8dd3c7", col = "white"))
				)[ind],
				grid_width = unit(0.6, "cm"),
				labels_gp = gpar(fontsize = legend_fontsize))
		}

		lgd2 = NULL
		mdf = x$df_imports[is_field_required(x$dep_fields), , drop = FALSE]
		ind = numeric(0)
		if(any(mdf[, 1] == 0 & mdf[, 2] == 0 & mdf[, 3] == 0)) {
			ind = c(ind, 1)
		}
		if(any(mdf[, 1] < 0 & is.finite(mdf[, 1]))) {
			ind = c(ind, 2)
		}
		if(any(is.infinite(mdf[, 1]))) {
			ind = c(ind, 3)
		}
		if(length(ind)) {
			lgd2 = Legend(title = "", at = c("The whole namespace is imported", "The whole namespace is imported except some functions", "Package is listed in 'Imports' but namespace is not imported")[ind], 
				type = "lines",
				legend_gp = gpar(lty = 2, col = c("red", "blue", "#808080")[ind]), grid_width = unit(0.6, "cm"), 
				labels_gp = gpar(fontsize = legend_fontsize))
		}

		if(!is.null(lgd1) && !is.null(lgd2)) {
			lgd_list = c(lgd_list, list(packLegend(lgd1, lgd2, row_gap = unit(2, "pt"), max_height = unit(10, "cm"))))
		} else if(!is.null(lgd1)) {
			lgd_list = c(lgd_list, list(lgd1))
		} else if(!is.null(lgd2)) {
			lgd_list = c(lgd_list, list(lgd2))
		}
	}

	ht = ht + rowAnnotation(pkg = anno_text(rownames(m), gp = gpar(fontsize = pkg_fontsize)))
	
	ht = draw(ht, 
		heatmap_legend_side = "bottom", 
		adjust_annotation_extension = FALSE,
		column_title = GetoptLong::qq("In total @{x$n_by_strong} packages are required directly or indirectly (@{x$n_by_all}) when installing '@{x$package}' (@{x$version})"),
		column_title_gp = gpar(fontsize = title_fontsize),
		heatmap_legend_list = lgd_list)

	w = convertWidth(getFromNamespace("width", ns = "ComplexHeatmap")(ht), "in", valueOnly = TRUE)
	h = convertHeight(getFromNamespace("height", ns = "ComplexHeatmap")(ht), "in", valueOnly = TRUE)

	if(fix_size && help) {
		ds = dev.size()
		if(w - ds[1] > 0.1 || h - ds[2] > 0.1) {
			message(paste0("The best device size to visualize the complete plot is ", 
				ceiling(w*100)/100, " x ", 
				ceiling(h*100)/100, " (in inches),\nor use `plot(obj, fix_size = FALSE)` so that heatmap cells are not in fixed sizes."))
		}
	}

	invisible(c(width = w, height = h))
}

scoreCol = function(x) {
	score = 0
	for(i in 1:length(x)) {
		if(!is.na(x[i])) {
			score = score + 2^(length(x)-i)
		}
	}
	return(score)
}

anno_nimports_barplot = function(x, category,
	bar_width = 0.6, ylim = NULL, axis = TRUE, gp = gpar(),
	axis_param = list(side = "top", labels_rot = 0),
	width = NULL, height = NULL, border = TRUE) {

	which = "row"

	anno_size = getFromNamespace("anno_width_and_height", ns = "ComplexHeatmap")(which, width, height, unit(2, "cm"))

	l = x[, 1] > 0 | x[, 2] > 0 | x[, 3] > 0
	if(!any(l) && is.null(ylim)) ylim = c(0, 1)

	if(any(l)) {
		data_scale = c(0, max(rowSums(x[l, , drop = FALSE])))
	} else {
		data_scale = c(0, 1)
	}

	if(is.null(ylim)) ylim = data_scale

	row_fun = function(index, k, n) {
		x2 = x
		x2[x2 < 0] = 0
		fun = anno_barplot(x2, gp = gp, which = "row", ylim = ylim,
			baseline = 0, width = anno_size$width, border = border, bar_width = bar_width,
			axis = axis, axis_param = axis_param)@fun
		fun(index, k, n)

		n = length(index)
		pushViewport(viewport(xscale = c(0, 1), yscale = c(0.5, n + 0.5)))
			
		if(is_field_required(category[index[1]])) {
			v = x[index, , drop = FALSE]
			l = v[, 1] == 0 & v[, 2] == 0 & v[, 3] == 0
			if(any(l)) {
				grid.segments(0, n - which(l) + 1, 1, n - which(l) + 1, gp = gpar(lty = 2, col = "red"), default.units = "native")
			}
			l = v[, 1] < 0 & is.finite(v[, 1])
			if(any(l)) {
				grid.segments(0, n - which(l) + 1, 1, n - which(l) + 1, gp = gpar(lty = 2, col = "blue"), default.units = "native")
			}
			l = is.infinite(v[, 1])
			if(any(l)) {
				grid.segments(0, n - which(l) + 1, 1, n - which(l) + 1, gp = gpar(lty = 2, col = "#808080"), default.units = "native")
			}
		}
		popViewport()
	}

	fun = row_fun

	anno = AnnotationFunction(
		fun = fun,
		fun_name = "anno_barplot",
		which = which,
		width = anno_size$width,
		height = anno_size$height,
		var_import = list(x, category, border, bar_width, axis, axis_param, anno_size, ylim, gp, is_field_required)
	)
	
	cl = new("AnnotationFunction")
	if("subsettable" %in% slotNames(cl)) {
		anno@subsettable = TRUE
	} else {
		anno@subsetable = FALSE
	}

	axis_param = getFromNamespace("validate_axis_param", ns = "ComplexHeatmap")(axis_param, which)
	axis_grob = if(axis) getFromNamespace("construct_axis_grob", ns = "ComplexHeatmap")(axis_param, which, data_scale) else NULL
	anno@extended = getFromNamespace("update_anno_extend", ns = "ComplexHeatmap")(anno, axis_grob, axis_param)

	return(anno) 
}

---
title: "Co-heaviness of two parent packages"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Co-heaviness of two parent packages}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r, echo = FALSE}
library(knitr)
knitr::opts_chunk$set(
    warning = FALSE,
    message = FALSE,
    fig.align = "center")
```

The heaviness measures the number of additional dependency packages that a
parent package uniquely imports, which are the dependency packages not imported by any
of the other parent packages. However, there are scenarios when multiple
parents import similar sets of dependencies, which results in heaviness for
individual parent being very small.

We take the **DESeq2** package (version 1.36.0. Note dependency has been optimized in recent versions of **DESeq2**) as an example. First we generate the dependency heatmap.

```{r, eval = FALSE}
library(pkgndep)
x = pkgndep("DESeq2")
```

```{r, echo = FALSE}
library(pkgndep)
x = readRDS(system.file("extdata", "DESeq2_dep.rds", package = "pkgndep"))
```

```{r, fig.width = 36.96, fig.height = 8.34, out.width="1200px"}
dependency_heatmap(x)
```

You can drag the plot into a new tab if it is too small to read.

```{r}
# or create the plot in an HTML page
# dependency_report(x)
```


As can be observed from the heatmap, **DESeq2**'s two parents **geneplotter**
and **genefilter** (the last two rows in "Imports" category in the heatmap) import 51
and 53 dependency packages, among them 50 packages are the same. Due to the
high overlap, the heaviness of **geneplotter** and **genefilter** on
**DESeq2** are only 1 and 2 respectively. However, if taking the two parents
together, i.e., by moving both parents to "Suggests" of **DESeq2**, 23 dependency packages
can be reduced.

Here we define **the co-heaviness** that measures the number of additional
dependency packages brought by two parent packages. Denote a package as _P_ and
its two strong parent packages as _A_ and _B_, 
i.e., parent packages in "Depends", "Imports" and "LinkingTo",
denote $S_A$ as the set of reduced dependency packages when only moving _A_ to
"Suggests" of _P_, denote $S_B$ as the set of reduced dependency packages when
only moving _B_ to "Suggests" of _P_, and denote $S_{AB}$ as the set of reduced dependency
packages when moving _A_ and _B_ together to "Suggests" of _P_, the co-heaviness of
_A_, _B_ on _P_ is calculatd as 

$$ \left | S_{AB} \setminus  \cup (S_A, S_B) \right | $$

which is the number of reduced package only caused by co-action of A and B. Symbol $A \setminus B$ is the set
of elements in _A_ not in _B_ and $|A|$ is the number of elements in set _A_.

Denote $h_A$ as the heaviness of _A_ (as a strong parent) on _P_, $h_B$ as the
heaviness of _B_ (as a strong parent) on _P_, and $h_{AB}$ as the co-heaviess
of _A_/_B_ on _P_, the number of reduced strong dependencies by moving both
_A_ and _B_ to _P_'s "Suggests" is:

$$ h_A + h_B + h_{AB} $$

**pkgndep** provides a function `co_heaviness()` that calculates co-heaviness of two parent packages.
It returns a co-heaviness matrix. Please note it only returns the matrix for strong parents.

```{r}
m = co_heaviness(x)
m
```

The co-heaviness matrix can be visualized as a heatmap.

```{r, fig.width = 6.5, fig.height = 5}
library(ComplexHeatmap)
Heatmap(m, name = "co-heaviness")
```

There are two major reasons for the high co-heaviness from two parent packages. 1. Parent _A_
also depends on package _B_, thus all the dependency packages brought by package _B_ will also be 
included in the dependency packages from package _A_. E.g., package **ffpe**'s two parents **lumi** and **methylumi** where **lumi** also depends on **methylumi**
and most of the heaviness comes from **methylummi**. 2. Although _A_ and _B_ do not depend on each other,
they have a common upstream package, which brings the same dependencies to A and B. Such as the example
of **DESeq2**, its two parent packages **geneplotter** and **genefilter** have a common upstream pacakge **annotate**
that contribute many dependencies to both package.

Alternatively, the co-heaviness can also be defined as a relative measure, the Jaccard coeffcient.
Note $S_{AB} \setminus  \cup (S_A, S_B)$ is actually the set of dependencies that are imported simultaneously by and only by two parent packages A and B. Thus the Jaccard coeffcient is calculated as:

$$ \frac{ \left | S_{AB} \setminus  \cup (S_A, S_B) \right | }{ \left | S_{AB} \right |  } $$

Or using the denotation of heaviness defined before, the Jaccard coeffcient can also be written as:

$$ \frac{h_{AB}}{h_A + h_B + h_{AB}} $$

The Jaccard coeffcient can be calculated by setting argument `jaccard = TRUE`:

```{r}
co_heaviness(x, jaccard = TRUE)
```

But users need to be careful that a small value of $|S_{AB}|$ may lead to a large Jaccard coeffcient value.

Finally, co-heaviness captures additional dependency packages from and only
from two parents, while it does work for more parents. However, the effect of
heaviness from multiple parents can always be easily observed from the
dependency heatmap.

---
title: "Suggestions for optimizing package dependencies"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Suggestions for optimizing package dependencies}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE, message = FALSE}
library(pkgndep)
```


## For the strong parents

Heaviness analysis provides hints for reducing the complexity of package
dependencies, but how to optimize depends on the specific use of parent
packages in the corresponding package. First we need to be aware of that there
are scenarios when reduction of heavy parents could not be performed: 

1. When a package extends the functionality of a heavy parents, the parent must be its strong parent.
2. A heavy parent provides core functionality to a package.
3. S4 methods or S4 classes are imported from a parent.

Now let's assume the dependencies from heavy parents can be reduced. We give
the following suggestions for reducing them:

1. Re-implement the functions that were imported from heavy parents. Take
   package **mapStats** (version 2.4) as an example, an extremely heavy parent **Hmisc** can
   be observed where **Hmisc** has a heaviness of 49 which is almost 60% of
   the total number of required packages of **mapStats**. If moving **Hmisc**
   to “Suggests” of **mapStats**, the number of strong dependencies can be
   reduced from 83 to 34. [The dependency heaviness
   analysis](https://pkgndep.github.io/prefix_m/mapStats_dependency_report.html)
   shows there is only one function imported from **Hmisc**. A deep inspection
   into the source code of **mapStats** reveals that only a function
   `capitalize()` from **Hmisc** is imported to **mapStats**. `capitalize()`
   is a simple function that only capitalizes the first letter of a string.
   The 49 additional dependencies imported from **Hmisc** can be avoided by
   simply reimplementing a function `capitalize()` by developer’s own.

2. The previous scenario is actually not very common. More common cases are
   when heavy parents only provide analysis that are not frequently used in
   the package, which we call it as "secondary analysis". We suggest to move
   heavy parent packages that are less used to "Suggests" and they are only
   loaded when the corresponding functions are called. Dependency packages
   listed in "Suggests" by default are not mandatory to be installed, thus,
   there will be errors when the functions are called but these dependency
   packages are not installed yet. We can write simple code to test whether 
   such secondary packages are installed. If not, print messages to tell users
   to install them.

3. Another way for reducing the package dependencies is to directly copy code
   from heavy parents. This approach is of course NOT recommended from the aspect
   of software engineering, [but actually it is widely used in CRAN
   packages](https://doi.org/10.1109/IWSC.2015.7069885).

4. Try to separate a large package into several smaller packages which focus on more specific tasks.
   For example, the package **remotes** separated from the package **devtools** only focuses on 
   installing packages from remote repositories, where [**remotes** only has 4 strong dependencies](https://pkgndep.github.io/prefix_r/remotes_dependency_report.html), 
   but [**devtools** has 76](https://pkgndep.github.io/prefix_d/devtools_dependency_report.html).

## For the weak parents

Weak parent packages are those listed in "Suggested/Enhances" fields in the
DESCRIPTION file. The heaviness mainly measures the effects on strong parents
(those in "Depends/Imports") and we suggest moving heavy and strong parents to
"Suggested/Enhances" if they are not frequently used by users. As "suggested"
packages are not necessarily required for users, optimizing the dependency
complexity for suggested packages is less necessary.

But still, on the developer's side, it is still important to keep the
dependency complexity of "all parent packages" to a reasonable size. CRAN and
Bioconductor perform full checks on packages, which means they require all the
packages in "Depends", "Imports" and "Suggests" to be installed. A large
dependency (normally from suggested packages) more likely produces failures,
which prevents acceptance or updates on CRAN or on Bioconductor. In general, we
have the following suggestions for handling dependency complexity from suggested
packages.

1. If removing a heavy suggested package does not affect the functionality of
   the package, or in other words, this dependency package only "enhances" the
   package, then it can be put in the "Enhances" field of the DESCRIPTION
   file. Packages listed in "Enhances" will never be checked by `R CMD check`.
2. If a heavy suggested package is only used in examples and in vignettes.
   Developers may think about whether they really need the corresponding code to run
   during the building and the checking of the package. More complex
   vignettes can be served separately while not being shipped with the package.
   A nice example is [the Seurat package](https://satijalab.org/seurat/).

---
title: "Check the heaviness of package dependencies"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Check the heaviness of package dependencies}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

When developing R packages, we should try to avoid directly setting
dependencies on "heavy packages". The "heaviness" for a package means, the
number of additional dependency packages it brings to. If your package directly depends
on a heavy package, it would bring several consequences:

1. Users need to install a lot of additional packages when installing your
   package which brings the risk that installation of some packages
   may fail and it makes your package cannot be installed. 
2. The namespaces that are loaded into your R session after loading your package will be huge (you can see the loaded namespaces by `sessionInfo()`).
3. You package will be "heavy" as well and it may take long time to load your package.

In the DESCRIPTION file of your package, there are "direct dependency
pakcages" listed in the `Depends`, `Imports` and `LinkingTo` fields. There are
also "indirect dependency packages" that can be found recursively for each of
the direct dependency packages. Here what we called "dependency packages" are
the union of the direct and indirect dependency packages.

There are also packages listed in `Suggests` and `Enhances` fields in
DESCRIPTION file, but they are not enforced to be installed when installing
your package. Of course, they also have "indirect dependency packages". To get
rid of the heavy packages that are not often used in your package, it is
better to move them into the `Suggests`/`Enhances` fields and to load/install
them only when they are needed.

Here the **pkgndep** package checks the heaviness of the dependency packages
of your package. For each package listed in the `Depends`, `Imports`,
`LinkingTo` and `Suggests`/`Enhances` fields in the DESCRIPTION file,
**pkgndep** checks how many additional packages your package requires. The
summary of the dependency is visualized by a customized heatmap.

As an example, I am developing a package called
[**cola**](https://github.com/jokergoo/cola/) which depends on [a lot of other
packages](https://github.com/jokergoo/cola/blob/master/DESCRIPTION).
The dependency heatmap looks like follows (please drag the figure to a new tab to see it in its actual size):

```{r, echo = FALSE}
library(pkgndep)
x = readRDS(system.file("extdata", "cola_dep.rds", package = "pkgndep"))
pdf(NULL)
size = dependency_heatmap(x, help = FALSE)
invisible(dev.off())
```

```{r, fig.width = size[1], fig.height = size[2], out.width = "1000px", echo = FALSE}
dependency_heatmap(x)
```

In the heatmap, rows are the packages listed in `Depends`, `Imports` and
`Suggests` fields, columns are the additional dependency packages required for
each row package. The barplots on the right show the number of required
package, the number of imported functions/methods/classes (parsed from
NAMESPACE file) and the quantitative measure "heaviness" (the definition of
heaviness will be introduced later).

We can see if all the packages are put in the `Depends` or `Imports` field
(i.e. movig all suggsted packages to `Imports`), in total `r x$n_by_all`
packages are required, which are really a lot. Actually some of the heavy
packages such as **WGCNA**, **clusterProfiler** and **ReactomePA** (the last
three packages in the heatmap rows) are not very frequently used in **cola**,
moving them to `Suggests` field and using them only when they are needed
greatly helps to reduce the dependencies of **cola**. Now the number of required
packages are reduced to only `r x$n_by_strong`.

## Usage

To use this package:

```r
library(pkgndep)
pkg = pkgndep("package-name")  # if the package is already installed
dependency_heatmap(pkg)
```

or

```r
pkg = pkgndep("path-of-the-package")  # if the package has not been installed yet
dependency_heatmap(pkg)
```

The value for `pkgndep()` should be 1. a CRAN/Bioconductor package, 2. an installed package, 3. a path of a local package, 4. URL of a GitHub repository.

Executable examples:

```{r}
library(pkgndep)
pkg = pkgndep("ComplexHeatmap")
pkg
```

`pkgndep()` first needs to retrieve package databases both from remote repositories and local libraries, as you
can see the message from above code. This only happens once and the database is internally
saved and re-used.

We can directly use `dependency_heatmap()` function to create the dependency heatmap:

```{r, echo = FALSE}
pdf(NULL)
size = dependency_heatmap(pkg, help = FALSE)
invisible(dev.off())
```

```{r, fig.width = size[1], fig.height = size[2], out.width = "1000px"}
dependency_heatmap(pkg)
```

You can set the `file` argument to directly save the image into a figure where the figure
size is automatically calculated. Supported image formats are `png`/`jpg`/`svg`/`pdf`.

```{r, eval = FALSE}
dependency_heatmap(pkg, file = "test.png")
```

`heaviness_report()` function can generate an HTML report for the dependency heaviness analysis on the package.

```{r, eval = FALSE}
heaviness_report(pkg)
```

## Heaviness

The heaviness of package dependency can be measured quantitatively. **pkgndep** provides two measures: the absolute
measure and the relative measure.

The heaviness of a dependency package is calculated as follows. If package _B_
is in the `Depends`/`Imports`/`LinkingTo` fields of package _A_, which means, package _B_ 
is directly required for package _A_, denote `v1` as the total number of packages for package _A_, 
and denote `v2` as the total number of required packages if moving package
_B_ to `Suggests` in package _A_ (which means, now _B_ is not enforced to be installed for package _A_). The
absolute measure of heaviness is simply `v1 - v2` and relative measure is `(v1 + a)/(v2 + a)` where `a` is a small constant, e.g. 10.
So here the absolute heaviness for package _B_ on package _A_ is the number of additional packages
that package _B_ uniquely brings in.

In the second scenario, if package _B_ is in the `Suggests`/`Enhances` fields of package
_A_, now `v2` is the total number of required packages if moving package _B_ to `Imports` in package _A_,
the absolute measure of heaviness is `v2 - v1` and relative measure is `(v2 + a)/(v1 + a)`.

The heaviness score can be calculated by the function `heaviness()`:

```{r}
heaviness(pkg)
heaviness(pkg, rel = TRUE)
```


## A fast version of `tools::package_dependencies()`

The package dependencies are based on "package database" which is normally retrieved by `available.packages()`.
In **tools** package, there is a `package_dependencies()` function that can be used to get a list of dependency packages.
In the following example code, we retrieve the dependency packages for package **ggplot2**.

```{r}
chooseCRANmirror(ind = 1) # choose the mirror fro RStudio
db = available.packages()
```


```{r}
system.time(p1 <- tools::package_dependencies("ggplot2", db = db, recursive = TRUE)[[1]])
```

In **pkgndep**, we implement a faster version of `package_dependencies()` function. First the database needs
to be reformatted by `reformat_db()` function. The returned variable `db2` is a reference class object and
its method `db2$package_dependencies()` can be used to retrieve dependency packages.

```{r}
db2 = reformat_db(db)
db2
system.time(p2 <- db2$package_dependencies("ggplot2", recursive = TRUE, simplify = TRUE))
```

`p1` and `p2` are actually identical:

```{r}
identical(sort(p1), sort(p2))
```


## Session info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgndep.R
\name{pkgndep}
\alias{pkgndep}
\title{Package dependency analysis}
\usage{
pkgndep(
  package,
  lib = .libPaths(),
  bioc_version = pkgndep_opt$bioc_version,
  verbose = TRUE
)
}
\arguments{
\item{package}{Package name. The value can be 1. a CRAN/Bioconductor package, 2. an installed package, 3. a path of a local package, 4. URL of a GitHub repository.}

\item{lib}{Local library. The value can be a character vector specifying the local library paths, or \code{NULL} which directly uses \code{\link[=.libPaths]{.libPaths()}}. A value of \code{NA} means not to use local libraries.}

\item{bioc_version}{Bioconductor version.}

\item{verbose}{Whether to show messages.}
}
\value{
A \code{pkgndep} object.
}
\description{
Package dependency analysis
}
\examples{
\dontrun{
x = pkgndep("ComplexHeatmap")
}
# The `x` variable generated by `pkgndep()` is already saved in this package.
x = readRDS(system.file("extdata", "ComplexHeatmap_dep.rds", package = "pkgndep"))
x
dependency_heatmap(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.pkgndep}
\alias{plot.pkgndep}
\alias{dependency_heatmap}
\title{Make the dependency heatmap}
\usage{
\method{plot}{pkgndep}(x, ...)

dependency_heatmap(
  x,
  pkg_fontsize = 10 * cex,
  title_fontsize = 12 * cex,
  legend_fontsize = 10 * cex,
  fix_size = !dev.interactive(),
  cex = 1,
  help = TRUE,
  file = NULL,
  res = 144
)
}
\arguments{
\item{x}{An object from \code{\link[=pkgndep]{pkgndep()}}.}

\item{...}{Other arguments.}

\item{pkg_fontsize}{Font size for the package names.}

\item{title_fontsize}{Font size for the title.}

\item{legend_fontsize}{Font size for the legends.}

\item{fix_size}{Should the rows and columns in the heatmap have fixed size?}

\item{cex}{A factor multiplicated to all font sizes.}

\item{help}{Whether to print help message?}

\item{file}{A path of the figure. The size of the figure is automatically calculated.}

\item{res}{Resolution of the figure (only for png and jpeg).}
}
\value{
A vector of two numeric values (in inches) that correspond to the width and height of the plot.
}
\description{
Make the dependency heatmap
}
\details{
If \code{fix_size} is set to \code{TRUE}. The size of the whole plot can be obtained by:

\if{html}{\out{<div class="sourceCode">}}\preformatted{size = dependency_heatmap(x, fix_size = TRUE)
}\if{html}{\out{</div>}}

where \code{size} is a numeric vector of length two which are the width and height of the whole heatmap.
If \code{file} argument is set, the size of the figure is automatically calculated.

If there are no dependency packages stored in \code{x}, \code{NULL} is returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgndep.R
\name{print.pkgndep}
\alias{print.pkgndep}
\title{Print method}
\usage{
\method{print}{pkgndep}(x, ...)
}
\arguments{
\item{x}{An object from \code{pkgndep}.}

\item{...}{Other arguments.}
}
\description{
Print method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db.R
\name{reformat_db}
\alias{reformat_db}
\title{Format the package database}
\usage{
reformat_db(db, version = NULL)
}
\arguments{
\item{db}{A data frame returned from \code{\link[utils:available.packages]{utils::available.packages()}} or \code{\link[utils:installed.packages]{utils::installed.packages()}}.}

\item{version}{Version of the database, a self-defined text.}
}
\value{
A \code{pkg_db} object. There are the following methods:
\itemize{
\item \code{pkg_db$get_meta(package, field = NULL)} \code{field} can take values in "Package", "Version" and "Repository".
\item \code{pkg_db$get_dependency_table(package)} Get the dependency table.
\item \code{pkg_db$get_rev_dependency_table(package)} Get the reverse dependency table.
\item \code{pkg_db$package_dependencies(package, recursive = FALSE, reverse = FALSE, which = "strong", simplify = FALSE)} All the arguments are the same as in \code{\link[tools:package_dependencies]{tools::package_dependencies()}}. Argument \code{simplify} controls whether to return a data frame or a simplied vector.
}
}
\description{
Format the package database
}
\details{
It reformats the data frame of the package database into a \code{pkg_db} object.
}
\examples{
\dontrun{
db = available.packages()
db2 = reformat_db(db)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data.R
\name{load_pkg_db}
\alias{load_pkg_db}
\title{Load package database}
\usage{
load_pkg_db(
  lib = .libPaths(),
  bioc_version = pkgndep_opt$bioc_version,
  verbose = TRUE
)
}
\arguments{
\item{lib}{Local library. The value can be a character vector specifying the local library paths, or \code{NULL} which directly uses \code{\link[=.libPaths]{.libPaths()}}. A value of \code{NA} means not to use local libraries.}

\item{bioc_version}{Biconductor version.}

\item{verbose}{Whether to print messages.}
}
\description{
Load package database
}
\details{
It loads the package database from CRAN/Bioconductor and locally installed packages.

The database object internaly is cached for repeated use of other functions in this package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/heaviness.R
\name{co_heaviness}
\alias{co_heaviness}
\title{Co-heaviness for pairs of parent packages}
\usage{
co_heaviness(x, rel = FALSE, a = 10, jaccard = FALSE)
}
\arguments{
\item{x}{An object returned by \code{\link[=pkgndep]{pkgndep()}}.}

\item{rel}{Whether to return the absolute measure or the relative measure.}

\item{a}{A constant added for calculating the relative measure.}

\item{jaccard}{Whether to return Jaccard coeffcient?}
}
\description{
Co-heaviness for pairs of parent packages
}
\details{
Denote a package as P and its two strong parent packages as A and B, i.e., parent packages in "Depends", "Imports" and "LinkingTo",
the co-heaviness for A and B is calculated as follows.

Denote S_A as the set of reduced dependency packages when only moving A to "Suggests" of P, and denote S_B as the set of reduced dependency
packages when only moving B to "Suggests" of P, denote S_AB as the set of reduced dependency packages when moving A and B together to "Suggests" of P,
the co-heaviness of A, B on P is calculatd as \code{length(setdiff(S_AB, union(S_A, S_B)))}, which is the number of reduced package only caused by co-action of A and B.

Note the co-heaviness is only calculated for parent packages in "Depends", "Imports" and "LinkingTo".

When \code{jaccard} is set to \code{TRUE}, the function returns jaccard coeffcient. \code{setdiff(S_AB, union(S_A, S_B))} is actually
the set of dependencies imported by and only by two parent packages A and B. Thus the jaccard coeffcient is calculated as
\code{length(setdiff(S_AB, union(S_A, S_B)))/length(S_AB)}.
}
\examples{
\dontrun{
# DESeq version 1.36.0, the dependencies have been changed in later versions.
x = readRDS(system.file("extdata", "DESeq2_dep.rds", package = "pkgndep"))
hm = co_heaviness(x)
ComplexHeatmap::Heatmap(hm)
co_heaviness(x, jaccard = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global.R
\name{pkgndep_opt}
\alias{pkgndep_opt}
\title{Global parameters for pkgndep}
\usage{
pkgndep_opt(..., RESET = FALSE, READ.ONLY = NULL, LOCAL = FALSE, ADD = FALSE)
}
\arguments{
\item{...}{Arguments for the parameters, see "details" section}

\item{RESET}{Reset to default values.}

\item{READ.ONLY}{Please ignore.}

\item{LOCAL}{Pllease ignore.}

\item{ADD}{Please ignore.}
}
\description{
Global parameters for pkgndep
}
\details{
There are following parameters:
\itemize{
\item \code{bioc_version} The bioconductor version. Use numeric version for devel branch.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{current_bioc_version}
\alias{current_bioc_version}
\title{Current Bioconductor Version}
\usage{
current_bioc_version()
}
\description{
Current Bioconductor Version
}
\details{
The version information is from https://bioconductor.org/config.yaml
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{heaviness_report}
\alias{heaviness_report}
\alias{dependency_report}
\title{HTML report for package dependency heaviness analysis}
\usage{
heaviness_report(pkg, file = NULL)

dependency_report(...)
}
\arguments{
\item{pkg}{An object from \code{\link[=pkgndep]{pkgndep()}}.}

\item{file}{The path of the html file. If it is not specified, the report will be automatically opened in the web browser.}

\item{...}{pass to \code{\link[=heaviness_report]{heaviness_report()}}.}
}
\value{
The path of the HTML file of the report.
}
\description{
HTML report for package dependency heaviness analysis
}
\examples{
if(interactive()) {
    x = readRDS(system.file("extdata", "ComplexHeatmap_dep.rds", package = "pkgndep"))
    heaviness_report(x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/heaviness.R
\name{heaviness}
\alias{heaviness}
\title{Heaviness from parent packages}
\usage{
heaviness(x, rel = FALSE, a = 10, only_strong_dep = FALSE)
}
\arguments{
\item{x}{An object returned by \code{\link[=pkgndep]{pkgndep()}}.}

\item{rel}{Whether to return the absolute measure or the relative measure.}

\item{a}{A constant added for calculating the relative measure.}

\item{only_strong_dep}{Whether to only return the heaviness for strong parents.}
}
\description{
Heaviness from parent packages
}
\details{
The heaviness from a parent package is calculated as follows: If package B is in the \code{Depends}/\code{Imports}/\code{LinkingTo} fields of package A,
which means, package B is necessary for package A, denote \code{v1} as the total numbers of packages required for package A,
and \code{v2} as the total number of required packages if moving package B to \code{Suggests} (which means, now B is not necessary for A).
The absolute measure is simply \code{v1 -  v2} and relative measure is \code{(v1 + a)/(v2 + a)}.

In the second scenario, if B is in the \code{Suggests}/\code{Enhances} fields of package A, now \code{v2} is the total number of required packages if moving
B to \code{Imports}, the absolute measure is \code{v2 - v1} and relative measure is \code{(v2 + a)/(v1 + a)}.
}
\examples{
x = readRDS(system.file("extdata", "ComplexHeatmap_dep.rds", package = "pkgndep"))
heaviness(x)
heaviness(x, rel = TRUE)
}
